# Copyright (C) 2015  IPA Project Contributors, see COPYING for license

from __future__ import print_function, absolute_import

import errno
import os

from configparser import ConfigParser

from ipaplatform.paths import paths
from ipapython.dn import DN
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec
# pylint: disable=relative-import
from custodia.message.kem import KEMKeysStore
from custodia.message.kem import KEY_USAGE_SIG, KEY_USAGE_ENC, KEY_USAGE_MAP
# pylint: enable=relative-import
from jwcrypto.common import json_decode, json_encode
from jwcrypto.common import base64url_encode
from jwcrypto.jwk import JWK
from ipaserver.secrets.common import iSecLdap
from binascii import unhexlify
import ldap


IPA_REL_BASE_DN = 'cn=custodia,cn=ipa,cn=etc'
IPA_KEYS_QUERY = '(&(ipaKeyUsage={usage:s})(memberPrincipal={princ:s}))'
IPA_CHECK_QUERY = '(cn=enc/{host:s})'
RFC5280_USAGE_MAP = {KEY_USAGE_SIG: 'digitalSignature',
                     KEY_USAGE_ENC: 'dataEncipherment'}


class KEMLdap(iSecLdap):

    @property
    def keysbase(self):
        return '%s,%s' % (IPA_REL_BASE_DN, self.basedn)

    def _encode_int(self, i):
        I = hex(i).rstrip("L").lstrip("0x")
        return base64url_encode(unhexlify((len(I) % 2) * '0' + I))

    def _parse_public_key(self, ipa_public_key):
        public_key = serialization.load_der_public_key(ipa_public_key,
                                                       default_backend())
        num = public_key.public_numbers()
        if isinstance(num, rsa.RSAPublicNumbers):
            return {'kty': 'RSA',
                    'e': self._encode_int(num.e),
                    'n': self._encode_int(num.n)}
        elif isinstance(num, ec.EllipticCurvePublicNumbers):
            if num.curve.name == 'secp256r1':
                curve = 'P-256'
            elif num.curve.name == 'secp384r1':
                curve = 'P-384'
            elif num.curve.name == 'secp521r1':
                curve = 'P-521'
            else:
                raise TypeError('Unsupported Elliptic Curve')
            return {'kty': 'EC',
                    'crv': curve,
                    'x': self._encode_int(num.x),
                    'y': self._encode_int(num.y)}
        else:
            raise TypeError('Unknown Public Key type')

    def get_key(self, usage, principal):
        conn = self.connect()
        scope = ldap.SCOPE_SUBTREE

        ldap_filter = self.build_filter(IPA_KEYS_QUERY,
                                        {'usage': RFC5280_USAGE_MAP[usage],
                                         'princ': principal})
        r = conn.search_s(self.keysbase, scope, ldap_filter)
        if len(r) != 1:
            raise ValueError("Incorrect number of results (%d) searching for "
                             "public key for %s" % (len(r), principal))
        ipa_public_key = r[0][1]['ipaPublicKey'][0]
        jwk = self._parse_public_key(ipa_public_key)
        jwk['use'] = KEY_USAGE_MAP[usage]
        return json_encode(jwk)

    def check_host_keys(self, host):
        conn = self.connect()
        scope = ldap.SCOPE_SUBTREE

        ldap_filter = self.build_filter(IPA_CHECK_QUERY, {'host': host})
        r = conn.search_s(self.keysbase, scope, ldap_filter)
        if not r:
            raise ValueError("No public keys were found for %s" % host)
        return True

    def _format_public_key(self, key):
        if isinstance(key, str):
            jwkey = json_decode(key)
            if 'kty' not in jwkey:
                raise ValueError('Invalid key, missing "kty" attribute')
            if jwkey['kty'] == 'RSA':
                pubnum = rsa.RSAPublicNumbers(jwkey['e'], jwkey['n'])
                pubkey = pubnum.public_key(default_backend())
            elif jwkey['kty'] == 'EC':
                if jwkey['crv'] == 'P-256':
                    curve = ec.SECP256R1
                elif jwkey['crv'] == 'P-384':
                    curve = ec.SECP384R1
                elif jwkey['crv'] == 'P-521':
                    curve = ec.SECP521R1
                else:
                    raise TypeError('Unsupported Elliptic Curve')
                pubnum = ec.EllipticCurvePublicNumbers(
                    jwkey['x'], jwkey['y'], curve)
                pubkey = pubnum.public_key(default_backend())
            else:
                raise ValueError('Unknown key type: %s' % jwkey['kty'])
        elif isinstance(key, rsa.RSAPublicKey):
            pubkey = key
        elif isinstance(key, ec.EllipticCurvePublicKey):
            pubkey = key
        else:
            raise TypeError('Unknown key type: %s' % type(key))

        return pubkey.public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo)

    def _get_dn(self, usage, principal):
        servicename, host = principal.split('@')[0].split('/')
        name = '%s/%s' % (KEY_USAGE_MAP[usage], host)
        service_rdn = ('cn', servicename) if servicename != 'host' else DN()
        return DN(('cn', name), service_rdn, self.keysbase)

    def set_key(self, usage, principal, key):
        """
        Write key for the host or service.

        Service keys are nested one level beneath the 'cn=custodia'
        container, in the 'cn=<servicename>' container; this allows
        fine-grained control over key management permissions for
        specific services.

        The container is assumed to exist.

        """
        public_key = self._format_public_key(key)
        dn = self._get_dn(usage, principal)
        conn = self.connect()
        try:
            mods = [('objectClass', [b'nsContainer',
                                     b'ipaKeyPolicy',
                                     b'ipaPublicKeyObject',
                                     b'groupOfPrincipals']),
                    ('cn', dn[0].value.encode('utf-8')),
                    ('ipaKeyUsage', RFC5280_USAGE_MAP[usage].encode('utf-8')),
                    ('memberPrincipal', principal.encode('utf-8')),
                    ('ipaPublicKey', public_key)]
            conn.add_s(str(dn), mods)
        except ldap.ALREADY_EXISTS:
            mods = [(ldap.MOD_REPLACE, 'ipaPublicKey', public_key)]
            conn.modify_s(str(dn), mods)

    def del_key(self, usage, principal):
        """Delete key for host or service

        :returns: DN of removed key or None when key was not found
        """
        dn = self._get_dn(usage, principal)
        conn = self.connect()
        try:
            conn.delete_s(str(dn))
        except ldap.NO_SUCH_OBJECT:
            return None
        else:
            return dn


def newServerKeys(path, keyid):
    skey = JWK(generate='RSA', use='sig', kid=keyid)
    ekey = JWK(generate='RSA', use='enc', kid=keyid)
    with open(path, 'w') as f:
        os.fchmod(f.fileno(), 0o600)
        os.fchown(f.fileno(), 0, 0)
        f.write('[%s,%s]' % (skey.export(), ekey.export()))
    return [skey.get_op_key('verify'), ekey.get_op_key('encrypt')]


class IPAKEMKeys(KEMKeysStore):
    """A KEM Keys Store.

    This is a store that holds public keys of registered
    clients allowed to use KEM messages. It takes the form
    of an authorizer merely for the purpose of attaching
    itself to a 'request' so that later on the KEM Parser
    can fetch the appropariate key to verify/decrypt an
    incoming request and make the payload available.

    The KEM Parser will actually perform additional
    authorization checks in this case.

    SimplePathAuthz is extended here as we want to attach the
    store only to requests on paths we are configured to
    manage.
    """

    def __init__(self, config=None, ipaconf=paths.IPA_DEFAULT_CONF):
        super(IPAKEMKeys, self).__init__(config)
        conf = ConfigParser()
        self.host = None
        self.realm = None
        self.ldap_uri = config.get('ldap_uri', None)
        if conf.read(ipaconf):
            self.host = conf.get('global', 'host')
            self.realm = conf.get('global', 'realm')
            if self.ldap_uri is None:
                self.ldap_uri = conf.get('global', 'ldap_uri', raw=True)

        self._server_keys = None

    def find_key(self, kid, usage):
        if kid is None:
            raise TypeError('Key ID is None, should be a SPN')
        conn = KEMLdap(self.ldap_uri)
        return conn.get_key(usage, kid)

    def generate_server_keys(self):
        self.generate_keys('host')

    def generate_keys(self, servicename):
        principal = '%s/%s@%s' % (servicename, self.host, self.realm)
        # Neutralize the key with read if any
        self._server_keys = None
        # Generate private key and store it
        pubkeys = newServerKeys(self.config['server_keys'], principal)
        # Store public key in LDAP
        ldapconn = KEMLdap(self.ldap_uri)
        ldapconn.set_key(KEY_USAGE_SIG, principal, pubkeys[0])
        ldapconn.set_key(KEY_USAGE_ENC, principal, pubkeys[1])

    def remove_server_keys_file(self):
        """Remove keys from disk

        The method does not fail when the file is missing.
        """
        try:
            os.unlink(self.config['server_keys'])
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            return False
        else:
            return True

    def remove_server_keys(self):
        """Remove keys from LDAP and disk
        """
        self.remove_keys('host')

    def remove_keys(self, servicename):
        """Remove keys from LDAP and disk
        """
        self.remove_server_keys_file()
        principal = '%s/%s@%s' % (servicename, self.host, self.realm)
        if self.ldap_uri is not None:
            ldapconn = KEMLdap(self.ldap_uri)
            ldapconn.del_key(KEY_USAGE_SIG, principal)
            ldapconn.del_key(KEY_USAGE_ENC, principal)

    @property
    def server_keys(self):
        if self._server_keys is None:
            with open(self.config['server_keys']) as f:
                jsonkeys = f.read()
            dictkeys = json_decode(jsonkeys)
            self._server_keys = (JWK(**dictkeys[KEY_USAGE_SIG]),
                                 JWK(**dictkeys[KEY_USAGE_ENC]))
        return self._server_keys


# Manual testing
if __name__ == '__main__':
    IKK = IPAKEMKeys({'paths': '/',
                      'server_keys': '/etc/ipa/custodia/server.keys'})
    IKK.generate_server_keys()
    print(('SIG', IKK.server_keys[0].export_public()))
    print(('ENC', IKK.server_keys[1].export_public()))
    print(IKK.find_key('host/%s@%s' % (IKK.host, IKK.realm),
                       usage=KEY_USAGE_SIG))
    print(IKK.find_key('host/%s@%s' % (IKK.host, IKK.realm),
                       usage=KEY_USAGE_ENC))



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Parameter system for command plugins.

A `Param` instance can be used to describe an argument or option that a command
takes, or an attribute that a command returns.  The `Param` base class is not
used directly, but there are many subclasses for specific Python data types
(like `Str` or `Int`) and specific properties (like `Password`).

To create a `Param` instance, you must always provide the parameter *name*,
which should be the LDAP attribute name if the parameter describes the attribute
of an LDAP entry.  For example, we could create an `Str` instance describing the user's last-name attribute like this:

>>> from ipalib import Str
>>> sn = Str('sn')
>>> sn.name
'sn'

When creating a `Param`, there are also a number of optional kwargs which
which can provide additional meta-data and functionality.  For example, every
parameter has a *cli_name*, the name used on the command-line-interface.  By
default the *cli_name* is the same as the *name*:

>>> sn.cli_name
'sn'

But often the LDAP attribute name isn't user friendly for the command-line, so
you can override this with the *cli_name* kwarg:

>>> sn = Str('sn', cli_name='last')
>>> sn.name
'sn'
>>> sn.cli_name
'last'

Note that the RPC interfaces (and the internal processing pipeline) always use
the parameter *name*, regardless of what the *cli_name* might be.

A `Param` also has two translatable kwargs: *label* and *doc*.  These must both
be `Gettext` instances.  They both default to a place-holder `FixMe` instance,
a subclass of `Gettext` used to mark a missing translatable string:

>>> sn.label
FixMe('sn')
>>> sn.doc
FixMe('sn')

The *label* is a short phrase describing the parameter.  It's used on the CLI
when interactively prompting for values, and as a label for form inputs in the
web-UI.  The *label* should start with an initial capital.  For example:

>>> from ipalib import _
>>> sn = Str('sn',
...     cli_name='last',
...     label=_('Last name'),
... )
>>> sn.label
Gettext('Last name', domain='ipa', localedir=None)

The *doc* is a longer description of the parameter.  It's used on the CLI when
displaying the help information for a command, and as extra instruction for a
form input on the web-UI.  By default the *doc* is the same as the *label*:

>>> sn.doc
Gettext('Last name', domain='ipa', localedir=None)

But you can override this with the *doc* kwarg.  Like the *label*, the *doc*
should also start with an initial capital and should not end with any
punctuation.  For example:

>>> sn = Str('sn',
...     cli_name='last',
...     label=_('Last name'),
...     doc=_("The user's last name"),
... )
>>> sn.doc
Gettext("The user's last name", domain='ipa', localedir=None)

Demonstration aside, you should always provide at least the *label* so the
various UIs are translatable.  Only provide the *doc* if the parameter needs
a more detailed description for clarity.
"""

import re
import decimal
import base64
import datetime
from xmlrpc.client import MAXINT, MININT

import six
from cryptography import x509 as crypto_x509

from ipalib.text import _ as ugettext
from ipalib.base import check_name
from ipalib.plugable import ReadOnly, lock
from ipalib.errors import ConversionError, RequirementError, ValidationError
from ipalib.errors import (
    PasswordMismatch, Base64DecodeError, CertificateFormatError,
    CertificateOperationError
)
from ipalib.constants import TYPE_ERROR, CALLABLE_ERROR, LDAP_GENERALIZED_TIME_FORMAT
from ipalib.text import Gettext, FixMe
from ipalib.util import json_serialize, validate_idna_domain
from ipalib.x509 import (
    load_der_x509_certificate, IPACertificate, default_backend)
from ipalib.util import strip_csr_header, apirepr
from ipapython import kerberos
from ipapython.dn import DN
from ipapython.dnsutil import DNSName


def _is_null(value):
    if value:
        return False
    elif isinstance(value, (int, float, decimal.Decimal)):
        # 0 is not NULL
        return False
    else:
        return True

if six.PY3:
    unicode = str


class DefaultFrom(ReadOnly):
    """
    Derive a default value from other supplied values.

    For example, say you wanted to create a default for the user's login from
    the user's first and last names. It could be implemented like this:

    >>> login = DefaultFrom(lambda first, last: first[0] + last)
    >>> login(first='John', last='Doe')
    'JDoe'

    If you do not explicitly provide keys when you create a `DefaultFrom`
    instance, the keys are implicitly derived from your callback by
    inspecting ``callback.func_code.co_varnames``. The keys are available
    through the ``DefaultFrom.keys`` instance attribute, like this:

    >>> login.keys
    ('first', 'last')

    The callback is available through the ``DefaultFrom.callback`` instance
    attribute, like this:

    >>> login.callback  # doctest:+ELLIPSIS
    <function <lambda> at 0x...>
    >>> login.callback.func_code.co_varnames  # The keys
    ('first', 'last')

    The keys can be explicitly provided as optional positional arguments after
    the callback. For example, this is equivalent to the ``login`` instance
    above:

    >>> login2 = DefaultFrom(lambda a, b: a[0] + b, 'first', 'last')
    >>> login2.keys
    ('first', 'last')
    >>> login2.callback.func_code.co_varnames  # Not the keys
    ('a', 'b')
    >>> login2(first='John', last='Doe')
    'JDoe'

    If any keys are missing when calling your `DefaultFrom` instance, your
    callback is not called and ``None`` is returned.  For example:

    >>> login(first='John', lastname='Doe') is None
    True
    >>> login() is None
    True

    Any additional keys are simply ignored, like this:

    >>> login(last='Doe', first='John', middle='Whatever')
    'JDoe'

    As above, because `DefaultFrom.__call__` takes only pure keyword
    arguments, they can be supplied in any order.

    Of course, the callback need not be a ``lambda`` expression. This third
    example is equivalent to both the ``login`` and ``login2`` instances
    above:

    >>> def get_login(first, last):
    ...     return first[0] + last
    ...
    >>> login3 = DefaultFrom(get_login)
    >>> login3.keys
    ('first', 'last')
    >>> login3.callback.func_code.co_varnames
    ('first', 'last')
    >>> login3(first='John', last='Doe')
    'JDoe'
    """

    def __init__(self, callback, *keys):
        """
        :param callback: The callable to call when all keys are present.
        :param keys: Optional keys used for source values.
        """
        if not callable(callback):
            raise TypeError(
                CALLABLE_ERROR % ('callback', callback, type(callback))
            )
        self.callback = callback
        if len(keys) == 0:
            fc = callback.__code__
            if fc.co_flags & 0x0c:
                raise ValueError("callback: variable-length argument list not allowed")
            self.keys = fc.co_varnames[:fc.co_argcount]
        else:
            self.keys = keys
        for key in self.keys:
            if type(key) is not str:
                raise TypeError(
                    TYPE_ERROR % ('keys', str, key, type(key))
                )
        lock(self)

    def __repr__(self):
        args = tuple(repr(k) for k in self.keys)
        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join(args)
        )

    def __call__(self, **kw):
        """
        Call the callback if all keys are present.

        If all keys are present, the callback is called and its return value is
        returned.  If any keys are missing, ``None`` is returned.

        :param kw: The keyword arguments.
        """
        vals = tuple(kw.get(k, None) for k in self.keys)
        if None in vals:
            return None
        try:
            return self.callback(*vals)
        except Exception:
            pass
        return None

    def __json__(self):
        return self.keys


def parse_param_spec(spec):
    """
    Parse shorthand ``spec`` into to ``(name, kw)``.

    The ``spec`` string determines the parameter name, whether the parameter is
    required, and whether the parameter is multivalue according the following
    syntax:

    ======  =====  ========  ==========
    Spec    Name   Required  Multivalue
    ======  =====  ========  ==========
    'var'   'var'  True      False
    'var?'  'var'  False     False
    'var*'  'var'  False     True
    'var+'  'var'  True      True
    ======  =====  ========  ==========

    For example,

    >>> parse_param_spec('login')
    ('login', {'required': True, 'multivalue': False})
    >>> parse_param_spec('gecos?')
    ('gecos', {'required': False, 'multivalue': False})
    >>> parse_param_spec('telephone_numbers*')
    ('telephone_numbers', {'required': False, 'multivalue': True})
    >>> parse_param_spec('group+')
    ('group', {'required': True, 'multivalue': True})

    :param spec: A spec string.
    """
    if type(spec) is not str:
        raise TypeError(
            TYPE_ERROR % ('spec', str, spec, type(spec))
        )
    _map = {
        '?': dict(required=False, multivalue=False),
        '*': dict(required=False, multivalue=True),
        '+': dict(required=True, multivalue=True),
    }
    end = spec[-1]
    if end in _map:
        return (spec[:-1], _map[end])
    return (spec, dict(required=True, multivalue=False))


__messages = set()

def _(message):
    __messages.add(message)
    return message


class Param(ReadOnly):
    """
    Base class for all parameters.

    Param attributes:
    =================
    The behavior of Param class and subclasses can be controlled using the
    following set of attributes:

      - cli_name: option name in CLI
      - cli_short_name: one character version of cli_name
      - deprecated_cli_aliases: deprecated CLI aliases
      - label: very short description of the parameter. This value is used in
        when the Command output is printed to CLI or in a Command help
      - doc: parameter long description used in help
      - required: the parameter is marked as required for given Command
      - multivalue: indicates if the attribute is multivalued
      - primary_key: Command's parameter primary key is used for unique
        identification of an LDAP object and for sorting
      - normalizer: a custom function for Param value normalization
      - default_from: a custom function for generating default values of
        parameter instance
      - autofill: by default, only `required` parameters get a default value
        from the default_from function. When autofill is enabled, optional
        attributes get the default value filled too
      - query: this attribute is controlled by framework. When the `query`
        is enabled, framework assumes that the value is only queried and not
        inserted in the LDAP. Validation is then relaxed - custom
        parameter validators are skipped and only basic class validators are
        executed to check the parameter value
      - attribute: this attribute is controlled by framework and enabled for
        all LDAP objects parameters (unless parameter has "virtual_attribute"
        flag). All parameters with enabled `attribute` are being encoded and
        placed to an entry passed to LDAP Create/Update calls
      - include: a list of contexts where this parameter should be included.
        `Param.use_in_context()` provides further information.
      - exclude: a list of contexts where this parameter should be excluded.
        `Param.use_in_context()` provides further information.
      - flags: there are several flags that can be used to further tune the
        parameter behavior:
            * no_display (Output parameters only): do not display the parameter
            * no_create: do not include the parameter for crud.Create based
              commands
            * no_update: do not include the parameter for crud.Update based
              commands
            * no_search: do not include the parameter for crud.Search based
              commands
            * no_option: this attribute is not displayed in the CLI, usually
              because there's a better way of setting it (for example, a
              separate command)
            * virtual_attribute: the parameter is not stored physically in the
              LDAP and thus attribute `attribute` is not enabled
            * suppress_empty (Output parameters only): do not display parameter
              value when empty
            * ask_create: CLI asks for parameter value even when the parameter
              is not `required`. Applied for all crud.Create based commands
            * ask_update: CLI asks for parameter value even when the parameter
              is not `required`. Applied for all crud.Update based commands
            * req_update: The parameter is `required` in all crud.Update based
              commands
            * nonempty: This is an internal flag; a required attribute should
              be used instead of it.
              The value of this parameter must not be empty, but it may
              not be given at all. All crud.Update commands automatically
              convert required parameters to `nonempty` ones, so the value
              can be unspecified (unchanged) but cannot be deleted.
            * optional_create: do not require the parameter for crud.Create
              based commands
            * allow_mod_for_managed_permission: permission-mod allows changing
              the parameter for managed permissions
      - hint: this attribute is currently not used
      - alwaysask: when enabled, CLI asks for parameter value even when the
        parameter is not `required`
      - sortorder: used to sort a list of parameters for Command. See
        `Command.finalize()` for further information
      - confirm: if password, ask for confirmation
    """

    # This is a dummy type so that most of the functionality of Param can be
    # unit tested directly without always creating a subclass; however, a real
    # (direct) subclass must *always* override this class attribute.
    # If multiple types are permitted, set `type` to the canonical type and
    # `allowed_types` to a tuple of all allowed types.
    type = type(None) # Ouch, this wont be very useful in the real world!

    # Subclasses should override this with something more specific:
    type_error = _('incorrect type')

    # _convert_scalar operates only on scalar values
    scalar_error = _('Only one value is allowed')

    password = False

    kwargs = (
        ('cli_name', str, None),
        ('cli_short_name', str, None),
        ('deprecated_cli_aliases', frozenset, frozenset()),
        ('label', (str, Gettext), None),
        ('doc', (str, Gettext), None),
        ('required', bool, True),
        ('multivalue', bool, False),
        ('primary_key', bool, False),
        ('normalizer', callable, None),
        ('default_from', DefaultFrom, None),
        ('autofill', bool, False),
        ('query', bool, False),
        ('attribute', bool, False),
        ('include', frozenset, None),
        ('exclude', frozenset, None),
        ('flags', frozenset, frozenset()),
        ('hint', (str, Gettext), None),
        ('alwaysask', bool, False),
        ('sortorder', int, 2), # see finalize()
        ('option_group', unicode, None),
        ('cli_metavar', str, None),
        ('no_convert', bool, False),
        ('deprecated', bool, False),
        ('confirm', bool, True),

        # The 'default' kwarg gets appended in Param.__init__():
        # ('default', self.type, None),
    )

    @property
    def allowed_types(self):
        """The allowed datatypes for this Param"""
        return (self.type,)

    def __init__(self, name, *rules, **kw):
        # Merge in kw from parse_param_spec():
        (name, kw_from_spec) = parse_param_spec(name)
        check_name(name)
        if 'required' not in kw:
            kw['required'] = kw_from_spec['required']
        if 'multivalue' not in kw:
            kw['multivalue'] = kw_from_spec['multivalue']

        # Add 'default' to self.kwargs
        if kw.get('multivalue', True):
            self.kwargs += (('default', tuple, None),)
        else:
            self.kwargs += (('default', self.type, None),)

        # Wrap 'default_from' in a DefaultFrom if not already:
        df = kw.get('default_from')
        if callable(df) and not isinstance(df, DefaultFrom):
            kw['default_from'] = DefaultFrom(df)

        # Perform type validation on kw:
        for (key, kind, default) in self.kwargs:
            value = kw.get(key)
            if value is not None:
                if kind in (tuple, frozenset):
                    if type(value) in (list, tuple, set, frozenset):
                        value = kind(value)
                    elif type(value) is str:
                        value = kind([value])
                if kind is callable and not callable(value):
                    raise TypeError(
                        CALLABLE_ERROR % (key, value, type(value))
                    )
                elif (isinstance(kind, (type, tuple)) and
                      not isinstance(value, kind)):
                    raise TypeError(
                        TYPE_ERROR % (key, kind, value, type(value))
                    )
                kw[key] = value
            elif key not in ('required', 'multivalue'):
                kw.pop(key, None)

        # We keep these values to use in __repr__():
        if kw['required']:
            if kw['multivalue']:
                self.param_spec = name + '+'
            else:
                self.param_spec = name
        else:
            if kw['multivalue']:
                self.param_spec = name + '*'
            else:
                self.param_spec = name + '?'
        self.__kw = dict(kw)
        del self.__kw['required']
        del self.__kw['multivalue']

        self.name = name
        self.nice = '%s(%r)' % (self.__class__.__name__, self.param_spec)

        # Make sure no unknown kw were given:
        assert all(isinstance(t, type) for t in self.allowed_types)
        if not set(t[0] for t in self.kwargs).issuperset(self.__kw):
            extra = set(kw) - set(t[0] for t in self.kwargs)
            raise TypeError(
                '%s: takes no such kwargs: %s' % (self.nice,
                    ', '.join(repr(k) for k in sorted(extra))
                )
            )

        # We keep this copy with merged values also to use when cloning:
        self.__clonekw = dict(kw)

        # Merge in default for 'cli_name', label, doc if not given:
        if kw.get('cli_name') is None:
            kw['cli_name'] = self.name

        if kw.get('cli_metavar') is None:
            kw['cli_metavar'] = self.__class__.__name__.upper()

        if kw.get('label') is None:
            kw['label'] = FixMe(self.name)

        if kw.get('doc') is None:
            kw['doc'] = kw['label']

        # Add in class rules:
        class_rules = []
        for (key, kind, default) in self.kwargs:
            value = kw.get(key, default)
            if hasattr(self, key):
                raise ValueError('kwarg %r conflicts with attribute on %s' % (
                    key, self.__class__.__name__)
                )
            setattr(self, key, value)
            rule_name = '_rule_%s' % key
            if value is not None and hasattr(self, rule_name):
                class_rules.append(getattr(self, rule_name))
        check_name(self.cli_name)

        # Check that only 'include' or 'exclude' was provided:
        if None not in (self.include, self.exclude):
            raise ValueError(
                '%s: cannot have both %s=%r and %s=%r' % (
                    self.nice,
                    'include', self.include,
                    'exclude', self.exclude,
                )
            )

        # Check that all the rules are callable
        self.class_rules = tuple(class_rules)
        self.rules = rules
        if self.query:
            # by definition a query enforces no class or parameter rules
            self.all_rules = ()
        else:
            self.all_rules = self.class_rules + self.rules
        for rule in self.all_rules:
            if not callable(rule):
                raise TypeError(
                    '%s: rules must be callable; got %r' % (self.nice, rule)
                )

        # Check that cli_short_name is only 1 character long:
        if not (self.cli_short_name is None or len(self.cli_short_name) == 1):
            raise ValueError(
                '%s: cli_short_name can only be a single character: %s' % (
                    self.nice, self.cli_short_name)
            )

        # And we're done.
        lock(self)

    def __repr__(self):
        """
        Return an expresion that could construct this `Param` instance.
        """
        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join(self.__repr_iter())
        )

    def __repr_iter(self):
        yield repr(self.param_spec)
        for rule in self.rules:
            yield rule.__name__
        for key in sorted(self.__kw):
            value = self.__kw[key]
            if callable(value) and hasattr(value, '__name__'):
                value = value.__name__
            elif isinstance(value, int):
                value = str(value)
            elif isinstance(value, (tuple, set, frozenset)):
                value = apirepr(list(value))
            elif key == 'cli_name':
                # always represented as native string
                value = repr(value)
            else:
                value = apirepr(value)
            yield '%s=%s' % (key, value)

    def __call__(self, value, **kw):
        """
        One stop shopping.
        """
        if _is_null(value):
            value = self.get_default(**kw)
        else:
            value = self.convert(self.normalize(value))
        return value

    def get_param_name(self):
        """
        Return the right name of an attribute depending on usage.

        Normally errors should use cli_name, our "friendly" name. When
        using the API directly or *attr return the real name.
        """
        name = self.cli_name
        if not name:
            name = self.name
        return name

    def kw(self):
        """
        Iterate through ``(key,value)`` for all kwargs passed to constructor.
        """
        for key in sorted(self.__kw):
            value = self.__kw[key]
            if callable(value) and hasattr(value, '__name__'):
                value = value.__name__
            yield (key, value)

    def use_in_context(self, env):
        """
        Return ``True`` if this parameter should be used in ``env.context``.

        If a parameter is created with niether the ``include`` nor the
        ``exclude`` kwarg, this method will always return ``True``.  For
        example:

        >>> from ipalib.config import Env
        >>> param = Param('my_param')
        >>> param.use_in_context(Env(context='foo'))
        True
        >>> param.use_in_context(Env(context='bar'))
        True

        If a parameter is created with an ``include`` kwarg, this method will
        only return ``True`` if ``env.context`` is in ``include``.  For example:

        >>> param = Param('my_param', include=['foo', 'whatever'])
        >>> param.include
        frozenset(['foo', 'whatever'])
        >>> param.use_in_context(Env(context='foo'))
        True
        >>> param.use_in_context(Env(context='bar'))
        False

        If a paremeter is created with an ``exclude`` kwarg, this method will
        only return ``True`` if ``env.context`` is not in ``exclude``.  For
        example:

        >>> param = Param('my_param', exclude=['foo', 'whatever'])
        >>> param.exclude
        frozenset(['foo', 'whatever'])
        >>> param.use_in_context(Env(context='foo'))
        False
        >>> param.use_in_context(Env(context='bar'))
        True

        Note that the ``include`` and ``exclude`` kwargs are mutually exclusive
        and that at most one can be suppelied to `Param.__init__()`.  For
        example:

        >>> param = Param('nope', include=['foo'], exclude=['bar'])
        Traceback (most recent call last):
          ...
        ValueError: Param('nope'): cannot have both include=frozenset(['foo']) and exclude=frozenset(['bar'])

        So that subclasses can add additional logic based on other environment
        variables, the entire `config.Env` instance is passed in rather than
        just the value of ``env.context``.
        """
        if self.include is not None:
            return (env.context in self.include)
        if self.exclude is not None:
            return (env.context not in self.exclude)
        return True

    def safe_value(self, value):
        """
        Return a value safe for logging.

        This is used so that sensitive values like passwords don't get logged.
        For example:

        >>> p = Password('my_password')
        >>> p.safe_value(u'This is my password')
        u'********'
        >>> p.safe_value(None) is None
        True

        >>> s = Str('my_str')
        >>> s.safe_value(u'Some arbitrary value')
        u'Some arbitrary value'
        """
        if self.password and value is not None:
            return u'********'
        return value

    def clone(self, **overrides):
        """
        Return a new `Param` instance similar to this one.
        """
        return self.clone_rename(self.name, **overrides)

    def clone_rename(self, name, **overrides):
        """
        Return a new `Param` instance similar to this one, but named differently
        """
        return self.clone_retype(name, self.__class__, **overrides)

    def clone_retype(self, name, klass, **overrides):
        """
        Return a new `Param` instance similar to this one, but of a different type
        """
        kw = dict(self.__clonekw)
        kw.update(overrides)
        return klass(name, *self.rules, **kw)

    def normalize(self, value):
        """
        Normalize ``value`` using normalizer callback.

        For example:

        >>> param = Param('telephone',
        ...     normalizer=lambda value: value.replace('.', '-')
        ... )
        >>> param.normalize(u'800.123.4567')
        u'800-123-4567'

        If this `Param` instance was created with a normalizer callback and
        ``value`` is a unicode instance, the normalizer callback is called and
        *its* return value is returned.

        On the other hand, if this `Param` instance was *not* created with a
        normalizer callback, if ``value`` is *not* a unicode instance, or if an
        exception is caught when calling the normalizer callback, ``value`` is
        returned unchanged.

        :param value: A proposed value for this parameter.
        """
        if self.multivalue:
            if type(value) not in (tuple, list):
                value = (value,)
        if self.multivalue:
            return tuple(
                self._normalize_scalar(v) for v in value
            )
        else:
            return self._normalize_scalar(value)

    def _normalize_scalar(self, value):
        """
        Normalize a scalar value.

        This method is called once for each value in a multivalue.
        """
        if self.normalizer is None:
            return value
        try:
            return self.normalizer(value)
        except Exception:
            return value

    def convert(self, value):
        """
        Convert ``value`` to the Python type required by this parameter.

        For example:

        >>> scalar = Str('my_scalar')
        >>> scalar.type
        <type 'unicode'>
        >>> scalar.convert(43.2)
        u'43.2'

        (Note that `Str` is a subclass of `Param`.)

        All non-numeric, non-boolean values which evaluate to False will be
        converted to None.  For example:

        >>> scalar.convert(u'') is None  # An empty string
        True
        >>> scalar.convert([]) is None  # An empty list
        True

        Likewise, they will be filtered out of a multivalue parameter.
        For example:

        >>> multi = Str('my_multi', multivalue=True)
        >>> multi.convert([1.5, '', 17, None, u'Hello'])
        (u'1.5', u'17', u'Hello')
        >>> multi.convert([None, u'']) is None  # Filters to an empty list
        True

        Lastly, multivalue parameters will always return a ``tuple`` (assuming
        they don't return ``None`` as in the last example above).  For example:

        >>> multi.convert(42)  # Called with a scalar value
        (u'42',)
        >>> multi.convert([0, 1])  # Called with a list value
        (u'0', u'1')

        Note that how values are converted (and from what types they will be
        converted) completely depends upon how a subclass implements its
        `Param._convert_scalar()` method.  For example, see
        `Str._convert_scalar()`.

        :param value: A proposed value for this parameter.
        """
        if not self.no_convert:
            convert = self._convert_scalar
        else:
            def convert(value):
                if isinstance(value, unicode):
                    return value
                return self._convert_scalar(value)

        if _is_null(value):
            return
        if self.multivalue:
            if type(value) not in (tuple, list):
                value = (value,)
            values = tuple(
                convert(v) for v in value if not _is_null(v)
            )
            if len(values) == 0:
                return
            return values
        return convert(value)

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        for t in self.allowed_types:
            if isinstance(value, t):
                return value

        raise ConversionError(name=self.name, error=ugettext(self.type_error))

    def validate(self, value, supplied=None):
        """
        Check validity of ``value``.

        :param value: A proposed value for this parameter.
        :param supplied: True if this parameter was supplied explicitly.
        """
        if value is None:
            if self.required or (supplied and 'nonempty' in self.flags):
                raise RequirementError(name=self.name)
            return
        if self.deprecated:
            raise ValidationError(name=self.get_param_name(),
                                  error=_('this option is deprecated'))
        if self.multivalue:
            if type(value) is not tuple:
                raise TypeError(
                    TYPE_ERROR % ('value', tuple, value, type(value))
                )
            if len(value) < 1:
                raise ValueError('value: empty tuple must be converted to None')
            for v in value:
                self._validate_scalar(v)
        else:
            self._validate_scalar(value)

    def _validate_scalar(self, value, index=None):
        for t in self.allowed_types:
            if isinstance(value, t):
                break
        else:
            raise TypeError(
                TYPE_ERROR % (self.name, self.type, value, type(value))
            )
        for rule in self.all_rules:
            error = rule(ugettext, value)
            if error is not None:
                raise ValidationError(name=self.get_param_name(), error=error)

    def get_default(self, **kw):
        """
        Return the static default or construct and return a dynamic default.

        (In these examples, we will use the `Str` and `Bytes` classes, which
        both subclass from `Param`.)

        The *default* static default is ``None``.  For example:

        >>> s = Str('my_str')
        >>> s.default is None
        True
        >>> s.get_default() is None
        True

        However, you can provide your own static default via the ``default``
        keyword argument when you create your `Param` instance.  For example:

        >>> s = Str('my_str', default=u'My Static Default')
        >>> s.default
        u'My Static Default'
        >>> s.get_default()
        u'My Static Default'

        If you need to generate a dynamic default from other supplied parameter
        values, provide a callback via the ``default_from`` keyword argument.
        This callback will be automatically wrapped in a `DefaultFrom` instance
        if it isn't one already (see the `DefaultFrom` class for all the gory
        details).  For example:

        >>> login = Str('login', default=u'my-static-login-default',
        ...     default_from=lambda first, last: (first[0] + last).lower(),
        ... )
        >>> isinstance(login.default_from, DefaultFrom)
        True
        >>> login.default_from.keys
        ('first', 'last')

        Then when all the keys needed by the `DefaultFrom` instance are present,
        the dynamic default is constructed and returned.  For example:

        >>> kw = dict(last=u'Doe', first=u'John')
        >>> login.get_default(**kw)
        u'jdoe'

        Or if any keys are missing, your *static* default is returned.
        For example:

        >>> kw = dict(first=u'John', department=u'Engineering')
        >>> login.get_default(**kw)
        u'my-static-login-default'
        """
        if self.default_from is not None:
            default = self.default_from(**kw)
            if default is not None:
                try:
                    return self.convert(self.normalize(default))
                except Exception:
                    pass
        return self.default

    def sort_key(self, value):
        return value

    def __json__(self):
        json_dict = {}
        for a, k, _d in self.kwargs:
            if k in (callable, DefaultFrom):
                continue
            elif isinstance(getattr(self, a), frozenset):
                json_dict[a] = [k for k in getattr(self, a, [])]
            else:
                val = getattr(self, a, '')
                if val is None:
                    # ignore 'not set' because lack of their presence is
                    # the information itself
                    continue
                json_dict[a] = json_serialize(val)

        json_dict['class'] = self.__class__.__name__
        json_dict['name'] = self.name
        json_dict['type'] = self.type.__name__

        return json_dict


class Bool(Param):
    """
    A parameter for boolean values (stored in the ``bool`` type).
    """

    type = bool
    type_error = _('must be True or False')

    # FIXME: This my quick hack to get some UI stuff working, change these defaults
    #   --jderose 2009-08-28
    kwargs = Param.kwargs + (
        ('truths', frozenset, frozenset([1, u'1', True, u'true', u'TRUE'])),
        ('falsehoods', frozenset, frozenset([0, u'0', False, u'false', u'FALSE'])),
    )

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        if type(value) in self.allowed_types:
            return value
        if isinstance(value, str):
            value = value.lower()
        if value in self.truths:
            return True
        if value in self.falsehoods:
            return False
        if type(value) in (tuple, list):
            raise ConversionError(name=self.name,
                                  error=ugettext(self.scalar_error))
        raise ConversionError(name=self.name, error=ugettext(self.type_error))


class Flag(Bool):
    """
    A boolean parameter that always gets filled in with a default value.

    This `Bool` subclass forces ``autofill=True`` in `Flag.__init__()`.  If no
    default is provided, it also fills in a default value of ``False``.
    Lastly, unlike the `Bool` class, the default must be either ``True`` or
    ``False`` and cannot be ``None``.

    For example:

    >>> flag = Flag('my_flag')
    >>> (flag.autofill, flag.default)
    (True, False)

    To have a default value of ``True``, create your `Flag` intance with
    ``default=True``.  For example:

    >>> flag = Flag('my_flag', default=True)
    >>> (flag.autofill, flag.default)
    (True, True)

    Also note that creating a `Flag` instance with ``autofill=False`` will have
    no effect.  For example:

    >>> flag = Flag('my_flag', autofill=False)
    >>> flag.autofill
    True
    """

    def __init__(self, name, *rules, **kw):
        kw['autofill'] = True
        if 'default' not in kw:
            kw['default'] = False
        if type(kw['default']) is not bool:
            default = kw['default']
            raise TypeError(
                TYPE_ERROR % ('default', bool, default, type(default))
            )
        super(Flag, self).__init__(name, *rules, **kw)


class Number(Param):
    """
    Base class for the `Int` and `Decimal` parameters.
    """

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        if type(value) in self.allowed_types:
            return value
        if type(value) in (unicode, float, int):
            try:
                return self.type(value)
            except ValueError:
                pass
        if type(value) in (tuple, list):
            raise ConversionError(name=self.name,
                                  error=ugettext(self.scalar_error))
        raise ConversionError(name=self.name, error=ugettext(self.type_error))


class Int(Number):
    """
    A parameter for integer values (stored in the ``int`` type).
    """

    type = int
    allowed_types = (int,)
    type_error = _('must be an integer')

    kwargs = Param.kwargs + (
        ('minvalue', int, int(MININT)),
        ('maxvalue', int, int(MAXINT)),
    )

    @staticmethod
    def convert_int(value):
        if type(value) in Int.allowed_types:
            return value

        if type(value) is float:
            return int(value)

        if type(value) is unicode:
            if u'.' in value:
                return int(float(value))
            if six.PY3 and re.match('0[0-9]+', value):
                # 0-prefixed octal format
                return int(value, 8)
            return int(value, 0)

        raise ValueError(value)

    def __init__(self, name, *rules, **kw):
        super(Int, self).__init__(name, *rules, **kw)

        if (self.minvalue > self.maxvalue) and (self.minvalue is not None and self.maxvalue is not None):
            raise ValueError(
                '%s: minvalue > maxvalue (minvalue=%r, maxvalue=%r)' % (
                    self.nice, self.minvalue, self.maxvalue)
            )

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        try:
            return Int.convert_int(value)
        except ValueError:
            raise ConversionError(name=self.get_param_name(),
                                  error=ugettext(self.type_error))

    def _rule_minvalue(self, _, value):
        """
        Check min constraint.
        """
        assert isinstance(value, int)
        if value < self.minvalue:
            return _('must be at least %(minvalue)d') % dict(
                minvalue=self.minvalue,
            )
        else:
            return None

    def _rule_maxvalue(self, _, value):
        """
        Check max constraint.
        """
        assert isinstance(value, int)
        if value > self.maxvalue:
            return _('can be at most %(maxvalue)d') % dict(
                maxvalue=self.maxvalue,
            )
        else:
            return None


class Decimal(Number):
    """
    A parameter for floating-point values (stored in the ``Decimal`` type).

    Python Decimal type helps overcome problems tied to plain "float" type,
    e.g. problem with representation or value comparison. In order to safely
    transfer the value over RPC libraries, it is being converted to string
    which is then converted back to Decimal number.
    """

    type = decimal.Decimal
    type_error = _('must be a decimal number')

    kwargs = Param.kwargs + (
        ('minvalue', decimal.Decimal, None),
        ('maxvalue', decimal.Decimal, None),
        # round Decimal to given precision
        ('precision', int, None),
        # when False, number is normalized to non-exponential form
        ('exponential', bool, False),
        # set of allowed decimal number classes
        ('numberclass', tuple, ('-Normal', '+Zero', '+Normal')),
    )

    def __init__(self, name, *rules, **kw):
        for kwparam in ('minvalue', 'maxvalue', 'default'):
            value = kw.get(kwparam)
            if value is None:
                continue
            if isinstance(value, (str, float)):
                try:
                    value = decimal.Decimal(value)
                except Exception as e:
                    raise ValueError(
                       '%s: cannot parse kwarg %s: %s' % (
                        name, kwparam, str(e)))
                kw[kwparam] = value

        super(Decimal, self).__init__(name, *rules, **kw)

        if (self.minvalue is not None and
                self.maxvalue is not None and
                self.minvalue > self.maxvalue):
            raise ValueError(
                '%s: minvalue > maxvalue (minvalue=%s, maxvalue=%s)' % (
                    self.nice, self.minvalue, self.maxvalue)
            )

        if self.precision is not None and self.precision < 0:
            raise ValueError('%s: precision must be at least 0' % self.nice)

    def _rule_minvalue(self, _, value):
        """
        Check min constraint.
        """
        assert type(value) is decimal.Decimal
        if value < self.minvalue:
            return _('must be at least %(minvalue)s') % dict(
                minvalue=self.minvalue,
            )
        else:
            return None

    def _rule_maxvalue(self, _, value):
        """
        Check max constraint.
        """
        assert type(value) is decimal.Decimal
        if value > self.maxvalue:
            return _('can be at most %(maxvalue)s') % dict(
                maxvalue=self.maxvalue,
            )
        else:
            return None

    def _enforce_numberclass(self, value):
        numberclass = value.number_class()
        if numberclass not in self.numberclass:
            raise ValidationError(name=self.get_param_name(),
                    error=_("number class '%(cls)s' is not included in a list "
                            "of allowed number classes: %(allowed)s") \
                            % dict(cls=numberclass,
                                   allowed=u', '.join(self.numberclass))
                )

    def _enforce_precision(self, value):
        assert type(value) is decimal.Decimal
        if self.precision is not None:
            quantize_exp = decimal.Decimal(10) ** -int(self.precision)
            try:
                value = value.quantize(quantize_exp)
            except decimal.DecimalException as e:
                raise ConversionError(name=self.get_param_name(),
                                      error=unicode(e))
        return value

    def _remove_exponent(self, value):
        assert type(value) is decimal.Decimal

        if not self.exponential:
            try:
                # adopted from http://docs.python.org/library/decimal.html
                value = value.quantize(decimal.Decimal(1)) \
                        if value == value.to_integral() \
                        else value.normalize()
            except decimal.DecimalException as e:
                raise ConversionError(name=self.get_param_name(),
                                      error=unicode(e))

        return value

    def _test_and_normalize(self, value):
        """
        This method is run in conversion and normalization methods to test
        that the Decimal number conforms to Parameter boundaries and then
        normalizes the value.
        """
        self._enforce_numberclass(value)
        value = self._remove_exponent(value)
        value = self._enforce_precision(value)
        return value

    def _convert_scalar(self, value, index=None):
        if isinstance(value, (str, float)):
            try:
                value = decimal.Decimal(value)
            except decimal.DecimalException as e:
                raise ConversionError(name=self.get_param_name(),
                                      error=unicode(e))

        if isinstance(value, decimal.Decimal):
            return self._test_and_normalize(value)

        return super(Decimal, self)._convert_scalar(value)

    def _normalize_scalar(self, value):
        if isinstance(value, decimal.Decimal):
            return self._test_and_normalize(value)

        return super(Decimal, self)._normalize_scalar(value)

class Data(Param):
    """
    Base class for the `Bytes` and `Str` parameters.

    Previously `Str` was as subclass of `Bytes`.  Now the common functionality
    has been split into this base class so that ``isinstance(foo, Bytes)`` wont
    be ``True`` when ``foo`` is actually an `Str` instance (which is confusing).
    """

    kwargs = Param.kwargs + (
        ('minlength', int, None),
        ('maxlength', int, None),
        ('length', int, None),
        ('pattern_errmsg', (str,), None),
    )

    re = None
    re_errmsg = None

    def __init__(self, name, *rules, **kw):
        super(Data, self).__init__(name, *rules, **kw)

        if not (
            self.length is None or
            (self.minlength is None and self.maxlength is None)
        ):
            raise ValueError(
                '%s: cannot mix length with minlength or maxlength' % self.nice
            )

        if self.minlength is not None and self.minlength < 1:
            raise ValueError(
                '%s: minlength must be >= 1; got %r' % (self.nice, self.minlength)
            )

        if self.maxlength is not None and self.maxlength < 1:
            raise ValueError(
                '%s: maxlength must be >= 1; got %r' % (self.nice, self.maxlength)
            )

        if None not in (self.minlength, self.maxlength):
            if self.minlength > self.maxlength:
                raise ValueError(
                    '%s: minlength > maxlength (minlength=%r, maxlength=%r)' % (
                        self.nice, self.minlength, self.maxlength)
                )
            elif self.minlength == self.maxlength:
                raise ValueError(
                    '%s: minlength == maxlength; use length=%d instead' % (
                        self.nice, self.minlength)
                )

    def _rule_pattern(self, _, value):
        """
        Check pattern (regex) contraint.
        """
        assert type(value) in self.allowed_types
        if self.re.match(value) is None:
            if self.re_errmsg:
                return self.re_errmsg % dict(pattern=self.pattern,)
            else:
                return _('must match pattern "%(pattern)s"') % dict(
                    pattern=self.pattern,
                )
        else:
            return None


class Bytes(Data):
    """
    A parameter for binary data (stored in the ``str`` type).

    This class is named *Bytes* instead of *Str* so it's aligned with the
    Python v3 ``(str, unicode) => (bytes, str)`` clean-up.  See:

        http://docs.python.org/3.0/whatsnew/3.0.html

    Also see the `Str` parameter.
    """

    type = bytes
    type_error = _('must be binary data')
    kwargs = Data.kwargs + (
        ('pattern', (bytes,), None),
    )

    def __init__(self, name, *rules, **kw):
        if kw.get('pattern', None) is None:
            self.re = None
        else:
            self.re = re.compile(kw['pattern'])
        self.re_errmsg = kw.get('pattern_errmsg', None)
        super(Bytes, self).__init__(name, *rules, **kw)

    def _rule_minlength(self, _, value):
        """
        Check minlength constraint.
        """
        assert type(value) is bytes
        if len(value) < self.minlength:
            return _('must be at least %(minlength)d bytes') % dict(
                minlength=self.minlength,
            )
        else:
            return None

    def _rule_maxlength(self, _, value):
        """
        Check maxlength constraint.
        """
        assert type(value) is bytes
        if len(value) > self.maxlength:
            return _('can be at most %(maxlength)d bytes') % dict(
                maxlength=self.maxlength,
            )
        else:
            return None

    def _rule_length(self, _, value):
        """
        Check length constraint.
        """
        assert type(value) is bytes
        if len(value) != self.length:
            return _('must be exactly %(length)d bytes') % dict(
                length=self.length,
            )
        else:
            return None

    def _convert_scalar(self, value, index=None):
        if isinstance(value, unicode):
            try:
                value = base64.b64decode(value)
            except (TypeError, ValueError) as e:
                raise Base64DecodeError(reason=str(e))
        return super(Bytes, self)._convert_scalar(value)


class Certificate(Param):
    type = crypto_x509.Certificate
    type_error = _('must be a certificate')
    allowed_types = (IPACertificate, bytes, unicode)

    def _convert_scalar(self, value, index=None):
        """
        :param value: either DER certificate or base64 encoded certificate
        :returns: bytes representing value converted to DER format
        """
        if isinstance(value, bytes):
            try:
                value = value.decode('ascii')
            except UnicodeDecodeError:
                # value is possibly a DER-encoded certificate
                pass

        if isinstance(value, unicode):
            # if we received unicodes right away or we got them after the
            # decoding, we will now try to receive DER-certificate
            try:
                value = base64.b64decode(value)
            except (TypeError, ValueError) as e:
                raise Base64DecodeError(reason=str(e))

        if isinstance(value, bytes):
            # we now only have either bytes or an IPACertificate object
            # if it's bytes, make it an IPACertificate object
            try:
                value = load_der_x509_certificate(value)
            except ValueError as e:
                raise CertificateFormatError(error=str(e))

        return super(Certificate, self)._convert_scalar(value)


class CertificateSigningRequest(Param):
    type = crypto_x509.CertificateSigningRequest
    type_error = _('must be a certificate signing request')
    allowed_types = (crypto_x509.CertificateSigningRequest, bytes, unicode)

    def __extract_der_from_input(self, value):
        """
        Tries to get the DER representation of whatever we receive as an input

        :param value:
            bytes instance containing something we hope is a certificate
            signing request
        :returns:
            base64-decoded representation of whatever we found in case input
            had been something else than DER or something which resembles
            DER, in which case we would just return input
        """
        try:
            value.decode('utf-8')
        except UnicodeDecodeError:
            # possibly DER-encoded CSR or something similar
            return value

        value = strip_csr_header(value)
        return base64.b64decode(value)

    def _convert_scalar(self, value, index=None):
        """
        :param value:
            either DER csr, base64-encoded csr or an object implementing the
            cryptography.CertificateSigningRequest interface
        :returns:
            an object with the cryptography.CertificateSigningRequest interface
        """
        if isinstance(value, unicode):
            try:
                value = value.encode('ascii')
            except UnicodeDecodeError:
                raise CertificateOperationError('not a valid CSR')

        if isinstance(value, bytes):
            # try to extract DER from whatever we got
            value = self.__extract_der_from_input(value)
            try:
                value = crypto_x509.load_der_x509_csr(
                    value, backend=default_backend())
            except ValueError as e:
                raise CertificateOperationError(
                    error=_("Failure decoding Certificate Signing Request:"
                            " %s") % e)

        return super(CertificateSigningRequest, self)._convert_scalar(value)


class Str(Data):
    """
    A parameter for Unicode text (stored in the ``unicode`` type).

    This class is named *Str* instead of *Unicode* so it's aligned with the
    Python v3 ``(str, unicode) => (bytes, str)`` clean-up.  See:

        http://docs.python.org/3.0/whatsnew/3.0.html

    Also see the `Bytes` parameter.
    """

    kwargs = Data.kwargs + (
        ('pattern', (str,), None),
        ('noextrawhitespace', bool, True),
    )

    type = unicode
    type_error = _('must be Unicode text')

    def __init__(self, name, *rules, **kw):
        if kw.get('pattern', None) is None:
            self.re = None
        else:
            self.re = re.compile(kw['pattern'], re.UNICODE)
        self.re_errmsg = kw.get('pattern_errmsg', None)
        super(Str, self).__init__(name, *rules, **kw)

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        if type(value) in self.allowed_types:
            return value
        if type(value) in (int, float, decimal.Decimal):
            return self.type(value)
        if type(value) in (tuple, list):
            raise ConversionError(name=self.name,
                                  error=ugettext(self.scalar_error))
        raise ConversionError(name=self.name, error=ugettext(self.type_error))

    def _rule_noextrawhitespace(self, _, value):
        """
        Do not allow leading/trailing spaces.
        """
        assert type(value) is unicode
        if self.noextrawhitespace is False:
            return None
        if len(value) != len(value.strip()):
            return _('Leading and trailing spaces are not allowed')
        else:
            return None

    def _rule_minlength(self, _, value):
        """
        Check minlength constraint.
        """
        assert type(value) is unicode
        if len(value) < self.minlength:
            return _('must be at least %(minlength)d characters') % dict(
                minlength=self.minlength,
            )
        else:
            return None

    def _rule_maxlength(self, _, value):
        """
        Check maxlength constraint.
        """
        assert type(value) is unicode
        if len(value) > self.maxlength:
            return _('can be at most %(maxlength)d characters') % dict(
                maxlength=self.maxlength,
            )
        else:
            return None

    def _rule_length(self, _, value):
        """
        Check length constraint.
        """
        assert type(value) is unicode
        if len(value) != self.length:
            return _('must be exactly %(length)d characters') % dict(
                length=self.length,
            )
        else:
            return None

    def sort_key(self, value):
        return value.lower()

class IA5Str(Str):
    """
    An IA5String per RFC 4517
    """

    def __init__(self, name, *rules, **kw):
        super(IA5Str, self).__init__(name, *rules, **kw)

    def _convert_scalar(self, value, index=None):
        if isinstance(value, str):
            for char in value:
                if ord(char) > 127:
                    raise ConversionError(name=self.get_param_name(),
                        error=_('The character %(char)r is not allowed.') %
                            dict(char=char,)
                    )
        return super(IA5Str, self)._convert_scalar(value)


class Password(Str):
    """
    A parameter for passwords (stored in the ``unicode`` type).
    """

    password = True

    def _convert_scalar(self, value, index=None):
        if isinstance(value, (tuple, list)) and len(value) == 2:
            (p1, p2) = value
            if p1 != p2:
                raise PasswordMismatch(name=self.name)
            value = p1
        return super(Password, self)._convert_scalar(value)


class Enum(Param):
    """
    Base class for parameters with enumerable values.
    """

    kwargs = Param.kwargs + (
        ('values', tuple, tuple()),
    )

    def __init__(self, name, *rules, **kw):
        kw['cli_metavar'] = str([str(v) for v in kw.get('values', tuple())])
        super(Enum, self).__init__(name, *rules, **kw)
        for (i, v) in enumerate(self.values):
            if type(v) not in self.allowed_types:
                n = '%s values[%d]' % (self.nice, i)
                raise TypeError(
                    TYPE_ERROR % (n, self.type, v, type(v))
                )

        if len(self.values) < 1:
            raise ValueError(
                '%s: list of values must not be empty' % self.nice)

    def _rule_values(self, _, value, **kw):
        if value not in self.values:
            if len(self.values) == 1:
                return _("must be '%(value)s'") % dict(value=self.values[0])
            else:
                values = u', '.join("'%s'" % value for value in self.values)
                return _('must be one of %(values)s') % dict(values=values)
        else:
            return None

class BytesEnum(Enum):
    """
    Enumerable for binary data (stored in the ``str`` type).
    """

    type = unicode


class StrEnum(Enum):
    """
    Enumerable for Unicode text (stored in the ``unicode`` type).

    For example:

    >>> enum = StrEnum('my_enum', values=(u'One', u'Two', u'Three'))
    >>> enum.validate(u'Two', 'cli') is None
    True
    >>> enum.validate(u'Four', 'cli')
    Traceback (most recent call last):
      ...
    ValidationError: invalid 'my_enum': must be one of 'One', 'Two', 'Three'
    """

    type = unicode


class IntEnum(Enum):
    """
    Enumerable for integer data (stored in the ``int`` type).
    """

    type = int
    allowed_types = (int,)
    type_error = Int.type_error

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        try:
            return Int.convert_int(value)
        except ValueError:
            raise ConversionError(name=self.get_param_name(),
                                  error=ugettext(self.type_error))


class Any(Param):
    """
    A parameter capable of holding values of any type. For internal use only.
    """

    type = object

    def _convert_scalar(self, value, index=None):
        return value

    def _validate_scalar(self, value, index=None):
        for rule in self.all_rules:
            error = rule(ugettext, value)
            if error is not None:
                raise ValidationError(name=self.name, error=error)


class File(Str):
    """Text file parameter type.

    Accepts file names and loads their content into the parameter value.
    """
    open_mode = 'r'
    kwargs = Data.kwargs + (
        # valid for CLI, other backends (e.g. webUI) can ignore this
        ('stdin_if_missing', bool, False),
        ('noextrawhitespace', bool, False),
    )


class BinaryFile(Bytes):
    """Binary file parameter type
    """
    open_mode = 'rb'
    kwargs = Data.kwargs + (
        # valid for CLI, other backends (e.g. webUI) can ignore this
        ('stdin_if_missing', bool, False),
        ('noextrawhitespace', bool, False),
    )


class DateTime(Param):
    """
    DateTime parameter type.

    Accepts LDAP Generalized time without in the following format:
       '%Y%m%d%H%M%SZ'

    Accepts subset of values defined by ISO 8601:
        '%Y-%m-%dT%H:%M:%SZ'
        '%Y-%m-%dT%H:%MZ'
        '%Y-%m-%dZ'

    Also accepts above formats using ' ' (space) as a separator instead of 'T'.

    Refer to the `man strftime` for the explanations for the %Y,%m,%d,%H.%M,%S.
    """

    accepted_formats = [LDAP_GENERALIZED_TIME_FORMAT,  # generalized time
                        '%Y-%m-%dT%H:%M:%SZ',  # ISO 8601, second precision
                        '%Y-%m-%dT%H:%MZ',     # ISO 8601, minute precision
                        '%Y-%m-%dZ',           # ISO 8601, date only
                        '%Y-%m-%d %H:%M:%SZ',  # non-ISO 8601, second precision
                        '%Y-%m-%d %H:%MZ']     # non-ISO 8601, minute precision


    type = datetime.datetime
    type_error = _('must be datetime value')

    def _convert_scalar(self, value, index=None):
        if isinstance(value, str):
            if value == u'now':
                time = datetime.datetime.utcnow()
                return time
            else:
                for date_format in self.accepted_formats:
                    try:
                        time = datetime.datetime.strptime(value, date_format)
                        return time
                    except ValueError:
                        pass

            # If we get here, the strptime call did not succeed for any
            # the accepted formats, therefore raise error

            error = (_("does not match any of accepted formats: ") +
                      (', '.join(self.accepted_formats)))

            raise ConversionError(name=self.get_param_name(),
                                  error=error)

        return super(DateTime, self)._convert_scalar(value)


class AccessTime(Str):
    """
    Access time parameter type.

    Accepts values conforming to generalizedTime as defined in RFC 4517
    section 3.3.13 without time zone information.
    """
    def _check_HHMM(self, t):
        if len(t) != 4:
            raise ValueError('HHMM must be exactly 4 characters long')
        if not t.isnumeric():
            raise ValueError('HHMM non-numeric')
        hh = int(t[0:2])
        if hh < 0 or hh > 23:
            raise ValueError('HH out of range')
        mm = int(t[2:4])
        if mm < 0 or mm > 59:
            raise ValueError('MM out of range')

    def _check_dotw(self, t):
        if t.isnumeric():
            value = int(t)
            if value < 1 or value > 7:
                raise ValueError('day of the week out of range')
        elif t not in ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'):
            raise ValueError('invalid day of the week')

    def _check_dotm(self, t, month_num=1, year=4):
        if not t.isnumeric():
            raise ValueError('day of the month non-numeric')
        value = int(t)
        if month_num in (1, 3, 5, 7, 8, 10, 12):
            if value < 1 or value > 31:
                raise ValueError('day of the month out of range')
        elif month_num in (4, 6, 9, 11):
            if value < 1 or value > 30:
                raise ValueError('day of the month out of range')
        elif month_num == 2:
            if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0):
                if value < 1 or value > 29:
                    raise ValueError('day of the month out of range')
            else:
                if value < 1 or value > 28:
                    raise ValueError('day of the month out of range')

    def _check_wotm(self, t):
        if not t.isnumeric():
            raise ValueError('week of the month non-numeric')
        value = int(t)
        if value < 1 or value > 6:
            raise ValueError('week of the month out of range')

    def _check_woty(self, t):
        if not t.isnumeric():
            raise ValueError('week of the year non-numeric')
        value = int(t)
        if value < 1 or value > 52:
            raise ValueError('week of the year out of range')

    def _check_doty(self, t):
        if not t.isnumeric():
            raise ValueError('day of the year non-numeric')
        value = int(t)
        if value < 1 or value > 365:
            raise ValueError('day of the year out of range')

    def _check_month_num(self, t):
        if not t.isnumeric():
            raise ValueError('month number non-numeric')
        value = int(t)
        if value < 1 or value > 12:
            raise ValueError('month number out of range')

    def _check_interval(self, t, check_func):
        intervals = t.split(',')
        for i in intervals:
            if not i:
                raise ValueError('invalid time range')
            values = i.split('-')
            if len(values) > 2:
                raise ValueError('invalid time range')
            for v in values:
                check_func(v)
            if len(values) == 2:
                if int(values[0]) > int(values[1]):
                    raise ValueError('invalid time range')

    def _check_W_spec(self, ts, index):
        if ts[index] != 'day':
            raise ValueError('invalid week specifier')
        index += 1
        self._check_interval(ts[index], self._check_dotw)
        return index

    def _check_M_spec(self, ts, index):
        if ts[index] == 'week':
            self._check_interval(ts[index + 1], self._check_wotm)
            index = self._check_W_spec(ts, index + 2)
        elif ts[index] == 'day':
            index += 1
            self._check_interval(ts[index], self._check_dotm)
        else:
            raise ValueError('invalid month specifier')
        return index

    def _check_Y_spec(self, ts, index):
        if ts[index] == 'month':
            index += 1
            self._check_interval(ts[index], self._check_month_num)
            index = self._check_M_spec(ts, index + 1)
        elif ts[index] == 'week':
            self._check_interval(ts[index + 1], self._check_woty)
            index = self._check_W_spec(ts, index + 2)
        elif ts[index] == 'day':
            index += 1
            self._check_interval(ts[index], self._check_doty)
        else:
            raise ValueError('invalid year specifier')
        return index

    def _check_generalized(self, t):
        assert type(t) is unicode
        if len(t) not in (10, 12, 14):
            raise ValueError('incomplete generalized time')
        if not t.isnumeric():
            raise ValueError('time non-numeric')
        # don't check year value, with time travel and all :)
        self._check_month_num(t[4:6])
        year_num = int(t[0:4])
        month_num = int(t[4:6])
        self._check_dotm(t[6:8], month_num, year_num)
        if len(t) >= 12:
            self._check_HHMM(t[8:12])
        else:
            self._check_HHMM('%s00' % t[8:10])
        if len(t) == 14:
            s = int(t[12:14])
            if s < 0 or s > 60:
                raise ValueError('seconds out of range')

    def _check(self, time):
        ts = time.split()
        if ts[0] == 'absolute':
            if len(ts) != 4:
                raise ValueError('invalid format, must be \'absolute generalizedTime ~ generalizedTime\'')
            self._check_generalized(ts[1])
            if ts[2] != '~':
                raise ValueError('invalid time range separator')
            self._check_generalized(ts[3])
            if int(ts[1]) >= int(ts[3]):
                raise ValueError('invalid time range')
        elif ts[0] == 'periodic':
            index = None
            if ts[1] == 'yearly':
                index = self._check_Y_spec(ts, 2)
            elif ts[1] == 'monthly':
                index = self._check_M_spec(ts, 2)
            elif ts[1] == 'weekly':
                index = self._check_W_spec(ts, 2)
            elif ts[1] == 'daily':
                index = 1
            if index is None:
                raise ValueError('period must be yearly, monthy or daily, got \'%s\'' % ts[1])
            self._check_interval(ts[index + 1], self._check_HHMM)
        else:
            raise ValueError('time neither absolute or periodic')

    def _rule_required(self, _, value):
        try:
            self._check(value)
        except ValueError as e:
            raise ValidationError(name=self.get_param_name(), error=e.args[0])
        except IndexError:
            raise ValidationError(
                name=self.get_param_name(), error=ugettext('incomplete time value')
            )


class DNParam(Param):
    type = DN

    def _convert_scalar(self, value, index=None):
        """
        Convert a single scalar value.
        """
        if type(value) in self.allowed_types:
            return value

        try:
            dn = DN(value)
        except Exception as e:
            raise ConversionError(name=self.get_param_name(),
                                  error=ugettext(e))
        return dn


def create_param(spec):
    """
    Create an `Str` instance from the shorthand ``spec``.

    This function allows you to create `Str` parameters (the most common) from
    a convenient shorthand that defines the parameter name, whether it is
    required, and whether it is multivalue.  (For the definition of the
    shorthand syntax, see the `parse_param_spec()` function.)

    If ``spec`` is an ``str`` instance, it will be used to create a new `Str`
    parameter, which will be returned.  For example:

    >>> s = create_param('hometown?')
    >>> s
    Str('hometown?')
    >>> (s.name, s.required, s.multivalue)
    ('hometown', False, False)

    On the other hand, if ``spec`` is already a `Param` instance, it is
    returned unchanged.  For example:

    >>> b = Bytes('cert')
    >>> create_param(b) is b
    True

    As a plugin author, you will not call this function directly (which would
    be no more convenient than simply creating the `Str` instance).  Instead,
    `frontend.Command` will call it for you when it evaluates the
    ``takes_args`` and ``takes_options`` attributes, and `frontend.Object`
    will call it for you when it evaluates the ``takes_params`` attribute.

    :param spec: A spec string or a `Param` instance.
    """
    if isinstance(spec, Param):
        return spec
    if type(spec) is not str:
        raise TypeError(
            TYPE_ERROR % ('spec', (str, Param), spec, type(spec))
        )
    return Str(spec)


class DNSNameParam(Param):
    """
    Domain name parameter type.

    :only_absolute a domain name has to be absolute
        (makes it absolute from unicode input)
    :only_relative a domain name has to be relative
    """
    type = DNSName
    type_error = _('must be DNS name')
    kwargs = Param.kwargs + (
        ('only_absolute', bool, False),
        ('only_relative', bool, False),
    )

    def __init__(self, name, *rules, **kw):
        super(DNSNameParam, self).__init__(name, *rules, **kw)
        if self.only_absolute and self.only_relative:
            raise ValueError('%s: cannot be both absolute and relative' %
                             self.nice)

    def _convert_scalar(self, value, index=None):
        if isinstance(value, unicode):
            try:
                validate_idna_domain(value)
            except ValueError as e:
                raise ConversionError(name=self.get_param_name(),
                                      error=unicode(e))
            value = DNSName(value)

            if self.only_absolute and not value.is_absolute():
                value = value.make_absolute()

        return super(DNSNameParam, self)._convert_scalar(value)

    def _rule_only_absolute(self, _, value):
        if self.only_absolute and not value.is_absolute():
            return _('must be absolute')
        else:
            return None

    def _rule_only_relative(self, _, value):
        if self.only_relative and value.is_absolute():
            return _('must be relative')
        else:
            return None


class Dict(Param):
    """
    A parameter for dictionary.
    """

    type = dict
    type_error = _("must be dictionary")


class Principal(Param):
    """
    Kerberos principal name
    """

    type = kerberos.Principal
    type_error = _('must be Kerberos principal')
    kwargs = Param.kwargs + (
        ('require_service', bool, False),
    )

    @property
    def allowed_types(self):
        return (self.type, unicode)

    def _convert_scalar(self, value, index=None):
        if isinstance(value, unicode):
            try:
                value = kerberos.Principal(value)
            except ValueError:
                raise ConversionError(
                    name=self.get_param_name(),
                    error=_("Malformed principal: '%(value)s'") % dict(
                        value=value))

        return super(Principal, self)._convert_scalar(value)

    def _rule_require_service(self, _, value):
        if self.require_service and not value.is_service:
            raise ValidationError(
                name=self.get_param_name(),
                error=_("Service principal is required")
            )



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `ipalib.frontend` module.
"""

# FIXME: Pylint errors
# pylint: disable=no-member
import pytest
import six

from ipatests.util import raises, read_only
from ipatests.util import ClassChecker, create_test_api
from ipatests.util import assert_equal
from ipalib.constants import TYPE_ERROR
from ipalib.base import NameSpace
from ipalib import frontend, backend, plugable, errors, parameters, config
from ipalib import output, messages
from ipalib.parameters import Str
from ipapython.version import API_VERSION

if six.PY3:
    unicode = str


pytestmark = pytest.mark.tier0


def test_RULE_FLAG():
    assert frontend.RULE_FLAG == 'validation_rule'


def test_rule():
    """
    Test the `ipalib.frontend.rule` function.
    """
    flag = frontend.RULE_FLAG
    rule = frontend.rule
    def my_func():
        pass
    assert not hasattr(my_func, flag)
    rule(my_func)
    assert getattr(my_func, flag) is True
    @rule
    def my_func2():
        pass
    assert getattr(my_func2, flag) is True


def test_is_rule():
    """
    Test the `ipalib.frontend.is_rule` function.
    """
    is_rule = frontend.is_rule
    flag = frontend.RULE_FLAG

    class no_call:
        def __init__(self, value):
            if value is not None:
                assert value in (True, False)
                setattr(self, flag, value)

    class call(no_call):
        def __call__(self):
            pass

    assert is_rule(call(True))
    assert not is_rule(no_call(True))
    assert not is_rule(call(False))
    assert not is_rule(call(None))


class test_HasParam(ClassChecker):
    """
    Test the `ipalib.frontend.Command` class.
    """

    _cls = frontend.HasParam

    def test_get_param_iterable(self):
        """
        Test the `ipalib.frontend.HasParam._get_param_iterable` method.
        """
        api = 'the api instance'
        class WithTuple(self.cls):
            takes_stuff = ('one', 'two')
        o = WithTuple(api)
        assert o._get_param_iterable('stuff') is WithTuple.takes_stuff

        junk = ('three', 'four')
        class WithCallable(self.cls):
            def takes_stuff(self):
                return junk
        o = WithCallable(api)
        assert o._get_param_iterable('stuff') is junk

        class WithParam(self.cls):
            takes_stuff = parameters.Str('five')
        o = WithParam(api)
        assert o._get_param_iterable('stuff') == (WithParam.takes_stuff,)

        class WithStr(self.cls):
            takes_stuff = 'six'
        o = WithStr(api)
        assert o._get_param_iterable('stuff') == ('six',)

        class Wrong(self.cls):
            takes_stuff = ['seven', 'eight']
        o = Wrong(api)
        e = raises(TypeError, o._get_param_iterable, 'stuff')
        assert str(e) == '%s.%s must be a tuple, callable, or spec; got %r' % (
            'Wrong', 'takes_stuff', Wrong.takes_stuff
        )

    def test_filter_param_by_context(self):
        """
        Test the `ipalib.frontend.HasParam._filter_param_by_context` method.
        """
        api = 'the api instance'
        class Example(self.cls):
            def get_stuff(self):
                return (
                    'one',  # Make sure create_param() is called for each spec
                    'two',
                    parameters.Str('three', include='cli'),
                    parameters.Str('four', exclude='server'),
                    parameters.Str('five', exclude=['whatever', 'cli']),
                )
        o = Example(api)

        # Test when env is None:
        params = list(o._filter_param_by_context('stuff'))
        assert list(p.name for p in params) == [
            'one', 'two', 'three', 'four', 'five'
        ]
        for p in params:
            assert type(p) is parameters.Str

        # Test when env.context == 'cli':
        cli = config.Env(context='cli')
        assert cli.context == 'cli'
        params = list(o._filter_param_by_context('stuff', cli))
        assert list(p.name for p in params) == ['one', 'two', 'three', 'four']
        for p in params:
            assert type(p) is parameters.Str

        # Test when env.context == 'server'
        server = config.Env(context='server')
        assert server.context == 'server'
        params = list(o._filter_param_by_context('stuff', server))
        assert list(p.name for p in params) == ['one', 'two', 'five']
        for p in params:
            assert type(p) is parameters.Str

        # Test with no get_stuff:
        class Missing(self.cls):
            pass
        o = Missing(api)
        gen = o._filter_param_by_context('stuff')
        e = raises(NotImplementedError, list, gen)
        assert str(e) == 'Missing.get_stuff()'

        # Test when get_stuff is not callable:
        class NotCallable(self.cls):
            get_stuff = ('one', 'two')
        o = NotCallable(api)
        gen = o._filter_param_by_context('stuff')
        e = raises(TypeError, list, gen)
        assert str(e) == '%s.%s must be a callable; got %r' % (
            'NotCallable', 'get_stuff', NotCallable.get_stuff
        )


class test_Command(ClassChecker):
    """
    Test the `ipalib.frontend.Command` class.
    """

    _cls = frontend.Command

    def get_subcls(self):
        """
        Return a standard subclass of `ipalib.frontend.Command`.
        """
        class Rule:
            def __init__(self, name):
                self.name = name

            def __call__(self, _, value):
                if value != self.name:
                    return _('must equal %r') % self.name
                else:
                    return None

        default_from = parameters.DefaultFrom(
                lambda arg: arg,
                'default_from'
        )
        normalizer = lambda value: value.lower()

        class example(self.cls):
            takes_options = (
                parameters.Str('option0', Rule('option0'),
                    normalizer=normalizer,
                    default_from=default_from,
                ),
                parameters.Str('option1', Rule('option1'),
                    normalizer=normalizer,
                    default_from=default_from,
                ),
            )
        return example

    def get_instance(self, args=tuple(), options=tuple()):
        """
        Helper method used to test args and options.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        class example(self.cls):
            takes_args = args
            takes_options = options
        o = example(api)
        o.finalize()
        return o

    def test_class(self):
        """
        Test the `ipalib.frontend.Command` class.
        """
        assert self.cls.takes_options == tuple()
        assert self.cls.takes_args == tuple()

    def test_get_args(self):
        """
        Test the `ipalib.frontend.Command.get_args` method.
        """
        api = 'the api instance'
        assert list(self.cls(api).get_args()) == []
        args = ('login', 'stuff')
        o = self.get_instance(args=args)
        assert tuple(o.get_args()) == args

    def test_get_options(self):
        """
        Test the `ipalib.frontend.Command.get_options` method.
        """
        api = 'the api instance'
        options = list(self.cls(api).get_options())
        assert len(options) == 1
        assert options[0].name == 'version'
        options = ('verbose', 'debug')
        o = self.get_instance(options=options)
        assert len(tuple(o.get_options())) == 3
        assert 'verbose' in tuple(o.get_options())
        assert 'debug' in tuple(o.get_options())

    def test_args(self):
        """
        Test the ``ipalib.frontend.Command.args`` instance attribute.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        o = self.cls(api)
        o.finalize()
        assert type(o.args) is NameSpace
        assert len(o.args) == 0
        args = ('destination', 'source?')
        ns = self.get_instance(args=args).args
        assert type(ns) is NameSpace
        assert len(ns) == len(args)
        assert list(ns) == ['destination', 'source']
        assert type(ns.destination) is parameters.Str
        assert type(ns.source) is parameters.Str
        assert ns.destination.required is True
        assert ns.destination.multivalue is False
        assert ns.source.required is False
        assert ns.source.multivalue is False

        # Test TypeError:
        if six.PY2:
            e = raises(TypeError, self.get_instance, args=(u'whatever',))
            assert str(e) == TYPE_ERROR % (
                'spec', (str, parameters.Param), u'whatever', unicode)
        else:
            e = raises(TypeError, self.get_instance, args=(b'whatever',))
            assert str(e) == TYPE_ERROR % (
                'spec', (str, parameters.Param), b'whatever', bytes)

        # Test ValueError, required after optional:
        e = raises(ValueError, self.get_instance, args=('arg1?', 'arg2'))
        assert str(e) == "arg2: required argument after optional in %s arguments ['arg1?', 'arg2']" % (self.get_instance().name)

         # Test ValueError, scalar after multivalue:
        e = raises(ValueError, self.get_instance, args=('arg1+', 'arg2'))
        assert str(e) == 'arg2: only final argument can be multivalue'

    def test_max_args(self):
        """
        Test the ``ipalib.frontend.Command.max_args`` instance attribute.
        """
        o = self.get_instance()
        assert o.max_args == 0
        o = self.get_instance(args=('one?',))
        assert o.max_args == 1
        o = self.get_instance(args=('one', 'two?'))
        assert o.max_args == 2
        o = self.get_instance(args=('one', 'multi+',))
        assert o.max_args is None
        o = self.get_instance(args=('one', 'multi*',))
        assert o.max_args is None

    def test_options(self):
        """
        Test the ``ipalib.frontend.Command.options`` instance attribute.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        o = self.cls(api)
        o.finalize()
        assert type(o.options) is NameSpace
        assert len(o.options) == 1
        options = ('target', 'files*')
        ns = self.get_instance(options=options).options
        assert type(ns) is NameSpace
        assert len(ns) == len(options) + 1
        assert list(ns) == ['target', 'files', 'version']
        assert type(ns.target) is parameters.Str
        assert type(ns.files) is parameters.Str
        assert ns.target.required is True
        assert ns.target.multivalue is False
        assert ns.files.required is False
        assert ns.files.multivalue is True

    def test_output(self):
        """
        Test the ``ipalib.frontend.Command.output`` instance attribute.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        inst = self.cls(api)
        inst.finalize()
        assert type(inst.output) is NameSpace
        assert list(inst.output) == ['result']
        assert type(inst.output.result) is output.Output

    def test_iter_output(self):
        """
        Test the ``ipalib.frontend.Command._iter_output`` instance attribute.
        """
        api = 'the api instance'
        class Example(self.cls):
            pass
        inst = Example(api)

        inst.has_output = tuple()
        assert list(inst._iter_output()) == []

        wrong = ['hello', 'world']
        inst.has_output = wrong
        e = raises(TypeError, list, inst._iter_output())
        assert str(e) == 'Example.has_output: need a %r; got a %r: %r' % (
            tuple, list, wrong
        )

        wrong = ('hello', 17)
        inst.has_output = wrong
        e = raises(TypeError, list, inst._iter_output())
        assert str(e) == 'Example.has_output[1]: need a %r; got a %r: %r' % (
            (str, output.Output), int, 17
        )

        okay = ('foo', output.Output('bar'), 'baz')
        inst.has_output = okay
        items = list(inst._iter_output())
        assert len(items) == 3
        assert list(o.name for o in items) == ['foo', 'bar', 'baz']
        for o in items:
            assert type(o) is output.Output

    def test_convert(self):
        """
        Test the `ipalib.frontend.Command.convert` method.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        kw = dict(
            option0=u'1.5',
            option1=u'7',
        )
        o = self.subcls(api)
        o.finalize()
        for (key, value) in o.convert(**kw).items():
            assert_equal(unicode(kw[key]), value)

    def test_normalize(self):
        """
        Test the `ipalib.frontend.Command.normalize` method.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        kw = dict(
            option0=u'OPTION0',
            option1=u'OPTION1',
        )
        norm = dict((k, v.lower()) for (k, v) in kw.items())
        sub = self.subcls(api)
        sub.finalize()
        assert sub.normalize(**kw) == norm

    def test_get_default(self):
        """
        Test the `ipalib.frontend.Command.get_default` method.
        """
        # FIXME: Add an updated unit tests for get_default()

    def test_default_from_chaining(self):
        """
        Test chaining of parameters through default_from.
        """
        class my_cmd(self.cls):
            takes_options = (
                Str('option0'),
                Str('option1', default_from=lambda option0: option0),
                Str('option2', default_from=lambda option1: option1),
            )

            def run(self, *args, **options):
                return dict(result=options)

        kw = dict(option0=u'some value')

        api, _home = create_test_api()
        api.finalize()
        o = my_cmd(api)
        o.finalize()
        e = o.get_default(**kw)  # pylint: disable=not-callable
        assert type(e) is dict
        assert 'option2' in e
        assert e['option2'] == u'some value'

    def test_validate(self):
        """
        Test the `ipalib.frontend.Command.validate` method.
        """
        class api:
            env = config.Env(context='cli')
            @staticmethod
            def is_production_mode():
                return False

        sub = self.subcls(api)
        sub.finalize()

        # Check with valid values
        okay = dict(
            option0=u'option0',
            option1=u'option1',
            another_option='some value',
            version=API_VERSION,
        )
        sub.validate(**okay)

        # Check with an invalid value
        fail = dict(okay)
        fail['option0'] = u'whatever'
        e = raises(errors.ValidationError, sub.validate, **fail)
        assert_equal(e.name, u'option0')
        assert_equal(e.error, u"must equal 'option0'")

        # Check with a missing required arg
        fail = dict(okay)
        fail.pop('option1')
        e = raises(errors.RequirementError, sub.validate, **fail)
        assert e.name == 'option1'

    def test_execute(self):
        """
        Test the `ipalib.frontend.Command.execute` method.
        """
        api = 'the api instance'
        o = self.cls(api)
        e = raises(NotImplementedError, o.execute)
        assert str(e) == 'Command.execute()'

    def test_args_options_2_params(self):
        """
        Test the `ipalib.frontend.Command.args_options_2_params` method.
        """

        # Test that ZeroArgumentError is raised:
        o = self.get_instance()
        e = raises(errors.ZeroArgumentError, o.args_options_2_params, 1)
        assert e.name == 'example'

        # Test that MaxArgumentError is raised (count=1)
        o = self.get_instance(args=('one?',))
        e = raises(errors.MaxArgumentError, o.args_options_2_params, 1, 2)
        assert e.name == 'example'
        assert e.count == 1
        assert str(e) == "command 'example' takes at most 1 argument"

        # Test that MaxArgumentError is raised (count=2)
        o = self.get_instance(args=('one', 'two?'))
        e = raises(errors.MaxArgumentError, o.args_options_2_params, 1, 2, 3)
        assert e.name == 'example'
        assert e.count == 2
        assert str(e) == "command 'example' takes at most 2 arguments"

        # Test that OptionError is raised when an extra option is given:
        o = self.get_instance()
        e = raises(errors.OptionError, o.args_options_2_params, bad_option=True)
        assert e.option == 'bad_option'

        # Test that OverlapError is raised:
        o = self.get_instance(args=('one', 'two'), options=('three', 'four'))
        e = raises(errors.OverlapError, o.args_options_2_params,
            1, 2, three=3, two=2, four=4, one=1)
        assert e.names == "['one', 'two']"

        # Test the permutations:
        o = self.get_instance(args=('one', 'two*'), options=('three', 'four'))
        mthd = o.args_options_2_params
        assert mthd() == dict()
        assert mthd(1) == dict(one=1)
        assert mthd(1, 2) == dict(one=1, two=(2,))
        assert mthd(1, 21, 22, 23) == dict(one=1, two=(21, 22, 23))
        assert mthd(1, (21, 22, 23)) == dict(one=1, two=(21, 22, 23))
        assert mthd(three=3, four=4) == dict(three=3, four=4)
        assert mthd(three=3, four=4, one=1, two=2) == \
            dict(one=1, two=2, three=3, four=4)
        assert mthd(1, 21, 22, 23, three=3, four=4) == \
            dict(one=1, two=(21, 22, 23), three=3, four=4)
        assert mthd(1, (21, 22, 23), three=3, four=4) == \
            dict(one=1, two=(21, 22, 23), three=3, four=4)

    def test_args_options_2_entry(self):
        """
        Test `ipalib.frontend.Command.args_options_2_entry` method.
        """
        class my_cmd(self.cls):
            takes_args = (
                parameters.Str('one', attribute=True),
                parameters.Str('two', attribute=False),
            )
            takes_options = (
                parameters.Str('three', attribute=True, multivalue=True),
                parameters.Str('four', attribute=True, multivalue=False),
            )

            def run(self, *args, **kw):
                return self.args_options_2_entry(*args, **kw)

        args = ('one', 'two')
        kw = dict(three=('three1', 'three2'), four='four')

        api, _home = create_test_api()
        api.finalize()
        o = my_cmd(api)
        o.finalize()
        e = o.run(*args, **kw)
        assert type(e) is dict
        assert 'one' in e
        assert 'two' not in e
        assert 'three' in e
        assert 'four' in e
        assert e['one'] == 'one'
        assert e['three'] == ['three1', 'three2']
        assert e['four'] == 'four'

    def test_params_2_args_options(self):
        """
        Test the `ipalib.frontend.Command.params_2_args_options` method.
        """
        o = self.get_instance(args='one', options='two')
        assert o.params_2_args_options() == ((), {})
        assert o.params_2_args_options(one=1) == ((1,), {})
        assert o.params_2_args_options(two=2) == ((), dict(two=2))
        assert o.params_2_args_options(two=2, one=1) == ((1,), dict(two=2))

    def test_run(self):
        """
        Test the `ipalib.frontend.Command.run` method.
        """
        class my_cmd(self.cls):
            def execute(self, *args, **kw):
                return ('execute', args, kw)

            def forward(self, *args, **kw):
                return ('forward', args, kw)

        args = ('Hello,', 'world,')
        kw = dict(how_are='you', on_this='fine day?', version=API_VERSION)

        # Test in server context:
        api, _home = create_test_api(in_server=True)
        api.finalize()
        o = my_cmd(api)
        if six.PY2:
            assert o.run.__func__ is self.cls.run.__func__
        else:
            assert o.run.__func__ is self.cls.run
        out = o.run(*args, **kw)
        assert ('execute', args, kw) == out

        # Test in non-server context
        api, _home = create_test_api(in_server=False)
        api.finalize()
        o = my_cmd(api)
        if six.PY2:
            assert o.run.__func__ is self.cls.run.__func__
        else:
            assert o.run.__func__ is self.cls.run
        assert ('forward', args, kw) == o.run(*args, **kw)

    def test_messages(self):
        """
        Test correct handling of messages
        """
        class TestMessage(messages.PublicMessage):
            type = 'info'
            format = 'This is a message.'
            errno = 1234

        class my_cmd(self.cls):
            def execute(self, *args, **kw):
                result = {'name': 'execute'}
                messages.add_message(kw['version'], result, TestMessage())
                return result

            def forward(self, *args, **kw):
                result = {'name': 'forward'}
                messages.add_message(kw['version'], result, TestMessage())
                return result

        args = ('Hello,', 'world,')
        kw = dict(how_are='you', on_this='fine day?', version=API_VERSION)

        expected = [TestMessage().to_dict()]

        # Test in server context:
        api, _home = create_test_api(in_server=True)
        api.finalize()
        o = my_cmd(api)
        if six.PY2:
            assert o.run.__func__ is self.cls.run.__func__
        else:
            assert o.run.__func__ is self.cls.run
        assert {'name': 'execute', 'messages': expected} == o.run(*args, **kw)

        # Test in non-server context
        api, _home = create_test_api(in_server=False)
        api.finalize()
        o = my_cmd(api)
        if six.PY2:
            assert o.run.__func__ is self.cls.run.__func__
        else:
            assert o.run.__func__ is self.cls.run
        assert {'name': 'forward', 'messages': expected} == o.run(*args, **kw)

    def test_validate_output_basic(self):
        """
        Test the `ipalib.frontend.Command.validate_output` method.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        class Example(self.cls):
            has_output = ('foo', 'bar', 'baz')

        inst = Example(api)
        inst.finalize()

        # Test with wrong type:
        wrong = ('foo', 'bar', 'baz')
        e = raises(TypeError, inst.validate_output, wrong)
        assert str(e) == '%s.validate_output(): need a %r; got a %r: %r' % (
            'Example', dict, tuple, wrong
        )

        # Test with a missing keys:
        wrong = dict(bar='hello')
        e = raises(ValueError, inst.validate_output, wrong)
        assert str(e) == '%s.validate_output(): missing keys %r in %r' % (
            'Example', ['baz', 'foo'], wrong
        )

        # Test with extra keys:
        wrong = dict(foo=1, bar=2, baz=3, fee=4, azz=5)
        e = raises(ValueError, inst.validate_output, wrong)
        assert str(e) == '%s.validate_output(): unexpected keys %r in %r' % (
            'Example', ['azz', 'fee'], wrong
        )

        # Test with different keys:
        wrong = dict(baz=1, xyzzy=2, quux=3)
        e = raises(ValueError, inst.validate_output, wrong)
        assert str(e) == '%s.validate_output(): missing keys %r in %r' % (
            'Example', ['bar', 'foo'], wrong
        ), str(e)

    def test_validate_output_per_type(self):
        """
        Test `ipalib.frontend.Command.validate_output` per-type validation.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False

        class Complex(self.cls):
            has_output = (
                output.Output('foo', int),
                output.Output('bar', list),
            )
        inst = Complex(api)
        inst.finalize()

        wrong = dict(foo=17.9, bar=[18])
        e = raises(TypeError, inst.validate_output, wrong)
        assert str(e) == '%s:\n  output[%r]: need (%r,); got %r: %r' % (
            'Complex.validate_output()', 'foo', int, float, 17.9
        )

        wrong = dict(foo=18, bar=17)
        e = raises(TypeError, inst.validate_output, wrong)
        assert str(e) == '%s:\n  output[%r]: need (%r,); got %r: %r' % (
            'Complex.validate_output()', 'bar', list, int, 17
        )

    def test_validate_output_nested(self):
        """
        Test `ipalib.frontend.Command.validate_output` nested validation.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False

        class Subclass(output.ListOfEntries):
            pass

        # Test nested validation:
        class nested(self.cls):
            has_output = (
                output.Output('hello', int),
                Subclass('world'),
            )
        inst = nested(api)
        inst.finalize()
        okay = dict(foo='bar')
        nope = ('aye', 'bee')

        wrong = dict(hello=18, world=[okay, nope, okay])
        e = raises(TypeError, inst.validate_output, wrong)
        assert str(e) == output.emsg % (
            'nested', 'Subclass', 'world', 1, dict, tuple, nope
        )

        wrong = dict(hello=18, world=[okay, okay, okay, okay, nope])
        e = raises(TypeError, inst.validate_output, wrong)
        assert str(e) == output.emsg % (
            'nested', 'Subclass', 'world', 4, dict, tuple, nope
        )

    def test_get_output_params(self):
        """
        Test the `ipalib.frontend.Command.get_output_params` method.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        class example(self.cls):
            has_output_params = (
                'one',
                'two',
                'three',
            )
            takes_args = (
                'foo',
            )
            takes_options = (
                Str('bar'),
                'baz',
            )

        inst = example(api)
        inst.finalize()
        assert list(inst.get_output_params()) == [
            'one', 'two', 'three'
        ]
        assert list(inst.output_params) == ['one', 'two', 'three']


class test_LocalOrRemote(ClassChecker):
    """
    Test the `ipalib.frontend.LocalOrRemote` class.
    """
    _cls = frontend.LocalOrRemote

    def test_init(self):
        """
        Test the `ipalib.frontend.LocalOrRemote.__init__` method.
        """
        class api:
            @staticmethod
            def is_production_mode():
                return False
        o = self.cls(api)
        o.finalize()
        assert list(o.args) == []
        assert list(o.options) == ['server', 'version']
        op = o.options.server
        assert op.required is False
        assert op.default is False

    def test_run(self):
        """
        Test the `ipalib.frontend.LocalOrRemote.run` method.
        """
        class example(self.cls):
            takes_args = 'key?'

            def forward(self, *args, **options):
                return dict(result=('forward', args, options))

            def execute(self, *args, **options):
                return dict(result=('execute', args, options))

        # Test when in_server=False:
        api, _home = create_test_api(in_server=False)
        api.add_plugin(example)
        api.finalize()
        cmd = api.Command.example
        assert cmd(version=u'2.47') == dict(
            result=('execute', (), dict(version=u'2.47'))
        )
        assert cmd(u'var', version=u'2.47') == dict(
            result=('execute', (u'var',), dict(version=u'2.47'))
        )
        assert cmd(server=True, version=u'2.47') == dict(
            result=('forward', (), dict(version=u'2.47', server=True))
        )
        assert cmd(u'var', server=True, version=u'2.47') == dict(
            result=('forward', (u'var',), dict(version=u'2.47', server=True))
        )

        # Test when in_server=True (should always call execute):
        api, _home = create_test_api(in_server=True)
        api.add_plugin(example)
        api.finalize()
        cmd = api.Command.example
        assert cmd(version=u'2.47') == dict(
            result=('execute', (), dict(version=u'2.47', server=False))
        )
        assert cmd(u'var', version=u'2.47') == dict(
            result=('execute', (u'var',), dict(version=u'2.47', server=False))
        )
        assert cmd(server=True, version=u'2.47') == dict(
            result=('execute', (), dict(version=u'2.47', server=True))
        )
        assert cmd(u'var', server=True, version=u'2.47') == dict(
            result=('execute', (u'var',), dict(version=u'2.47', server=True))
        )


class test_Object(ClassChecker):
    """
    Test the `ipalib.frontend.Object` class.
    """
    _cls = frontend.Object

    def test_class(self):
        """
        Test the `ipalib.frontend.Object` class.
        """
        assert self.cls.backend is None
        assert self.cls.methods is None
        assert self.cls.params is None
        assert self.cls.params_minus_pk is None
        assert self.cls.takes_params == tuple()

    def test_init(self):
        """
        Test the `ipalib.frontend.Object.__init__` method.
        """
        # Setup for test:
        class DummyAttribute:
            def __init__(self, obj_name, attr_name, name=None):
                self.obj_name = obj_name
                self.attr_name = attr_name
                if name is None:
                    self.name = '%s_%s' % (obj_name, attr_name)
                else:
                    self.name = name
                self.bases = (DummyAttribute,)
                self.version = '1'
                self.full_name = '{}/{}'.format(self.name, self.version)
                self.param = frontend.create_param(attr_name)

            def __clone__(self, attr_name):
                return self.__class__(
                    self.obj_name,
                    self.attr_name,
                    getattr(self, attr_name)
                )

        def get_attributes(cnt, format):
            for name in ['other', 'user', 'another']:
                for i in range(cnt):
                    yield DummyAttribute(name, format % i)

        cnt = 10
        methods_format = 'method_%d'

        class FakeAPI:
            def __init__(self):
                self._API__plugins = get_attributes(cnt, methods_format)
                self._API__default_map = {}
                self.Method = plugable.APINameSpace(self, DummyAttribute)
            def __contains__(self, key):
                return hasattr(self, key)
            def __getitem__(self, key):
                return getattr(self, key)
            def is_production_mode(self):
                return False
            def _get(self, plugin):
                return plugin
        api = FakeAPI()
        assert len(api.Method) == cnt * 3

        class user(self.cls):
            pass

        # Actually perform test:
        o = user(api)
        assert read_only(o, 'api') is api

        namespace = o.methods
        assert isinstance(namespace, NameSpace)
        assert len(namespace) == cnt
        f = methods_format
        for i in range(cnt):
            attr_name = f % i
            attr = namespace[attr_name]
            assert isinstance(attr, DummyAttribute)
            assert attr is getattr(namespace, attr_name)
            assert attr.obj_name == 'user'
            assert attr.attr_name == attr_name
            assert attr.name == '%s_%s' % ('user', attr_name)

        # Test params instance attribute
        o = self.cls(api)
        ns = o.params
        assert type(ns) is NameSpace
        assert len(ns) == 0
        class example(self.cls):
            takes_params = ('banana', 'apple')
        o = example(api)
        ns = o.params
        assert type(ns) is NameSpace
        assert len(ns) == 2, repr(ns)
        assert list(ns) == ['banana', 'apple']
        for p in ns():
            assert type(p) is parameters.Str
            assert p.required is True
            assert p.multivalue is False

    def test_primary_key(self):
        """
        Test the `ipalib.frontend.Object.primary_key` attribute.
        """
        api, _home = create_test_api()
        api.finalize()

        # Test with no primary keys:
        class example1(self.cls):
            takes_params = (
                'one',
                'two',
            )
        o = example1(api)
        assert o.primary_key is None

        # Test with 1 primary key:
        class example2(self.cls):
            takes_params = (
                'one',
                'two',
                parameters.Str('three', primary_key=True),
                'four',
            )
        o = example2(api)
        pk = o.primary_key
        assert type(pk) is parameters.Str
        assert pk.name == 'three'
        assert pk.primary_key is True
        assert o.params[2] is o.primary_key
        assert isinstance(o.params_minus_pk, NameSpace)
        assert list(o.params_minus_pk) == ['one', 'two', 'four']

        # Test with multiple primary_key:
        class example3(self.cls):
            takes_params = (
                parameters.Str('one', primary_key=True),
                parameters.Str('two', primary_key=True),
                'three',
                parameters.Str('four', primary_key=True),
            )
        o = example3(api)
        e = raises(ValueError, o.finalize)
        assert str(e) == \
            'example3 (Object) has multiple primary keys: one, two, four'

    def test_backend(self):
        """
        Test the `ipalib.frontend.Object.backend` attribute.
        """
        api, _home = create_test_api()
        class ldap(backend.Backend):
            whatever = 'It worked!'
        api.add_plugin(ldap)
        class user(frontend.Object):
            backend_name = 'ldap'
        api.add_plugin(user)
        api.finalize()
        b = api.Object.user.backend
        assert isinstance(b, ldap)
        assert b.whatever == 'It worked!'

    def test_get_dn(self):
        """
        Test the `ipalib.frontend.Object.get_dn` method.
        """
        api = 'the api instance'
        o = self.cls(api)
        e = raises(NotImplementedError, o.get_dn, 'primary key')
        assert str(e) == 'Object.get_dn()'
        class user(self.cls):
            pass
        o = user(api)
        e = raises(NotImplementedError, o.get_dn, 'primary key')
        assert str(e) == 'user.get_dn()'

    def test_params_minus(self):
        """
        Test the `ipalib.frontend.Object.params_minus` method.
        """
        class example(self.cls):
            takes_params = ('one', 'two', 'three', 'four')
        api, _home = create_test_api()
        api.finalize()
        o = example(api)
        p = o.params
        assert tuple(o.params_minus()) == tuple(p())
        assert tuple(o.params_minus([])) == tuple(p())
        assert tuple(o.params_minus('two', 'three')) == (p.one, p.four)
        assert tuple(o.params_minus(['two', 'three'])) == (p.one, p.four)
        assert tuple(o.params_minus(p.two, p.three)) == (p.one, p.four)
        assert tuple(o.params_minus([p.two, p.three])) == (p.one, p.four)
        ns = NameSpace([p.two, p.three])
        assert tuple(o.params_minus(ns)) == (p.one, p.four)


class test_Attribute(ClassChecker):
    """
    Test the `ipalib.frontend.Attribute` class.
    """
    _cls = frontend.Attribute

    def test_class(self):
        """
        Test the `ipalib.frontend.Attribute` class.
        """
        assert self.cls.__bases__ == (plugable.Plugin,)
        assert type(self.cls.obj) is property
        assert type(self.cls.obj_name) is property
        assert type(self.cls.attr_name) is property

    def test_init(self):
        """
        Test the `ipalib.frontend.Attribute.__init__` method.
        """
        user_obj = 'The user frontend.Object instance'

        class api:
            Object = {("user", "1"): user_obj}
            @staticmethod
            def is_production_mode():
                return False

        class user_add(self.cls):
            pass

        o = user_add(api)
        assert read_only(o, 'api') is api
        assert read_only(o, 'obj') is user_obj
        assert read_only(o, 'obj_name') == 'user'
        assert read_only(o, 'attr_name') == 'add'


class test_Method(ClassChecker):
    """
    Test the `ipalib.frontend.Method` class.
    """
    _cls = frontend.Method

    def get_api(self, args=tuple(), options=tuple()):
        """
        Return a finalized `ipalib.plugable.API` instance.
        """
        api, _home = create_test_api()
        class user(frontend.Object):
            takes_params = (
                'givenname',
                'sn',
                frontend.Param('uid', primary_key=True),
                'initials',
            )
        class user_verb(self.cls):
            takes_args = args
            takes_options = options
        api.add_plugin(user)
        api.add_plugin(user_verb)
        api.finalize()
        return api

    def test_class(self):
        """
        Test the `ipalib.frontend.Method` class.
        """
        assert self.cls.__bases__ == (frontend.Attribute, frontend.Command)

    def test_init(self):
        """
        Test the `ipalib.frontend.Method.__init__` method.
        """
        api = 'the api instance'
        class user_add(self.cls):
            pass
        o = user_add(api)
        assert o.name == 'user_add'
        assert o.obj_name == 'user'
        assert o.attr_name == 'add'



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
YubiKey Tokens

Manage YubiKey tokens.

This code is an extension to the otptoken plugin and provides support for
reading/writing YubiKey tokens directly.

EXAMPLES:

 Add a new token:
   ipa otptoken-add-yubikey --owner=jdoe --desc="My YubiKey"
""")

register = Registry()



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#
from __future__ import absolute_import

import os

from ipaplatform.tasks import tasks


def test_ipa_version():
    v3 = tasks.parse_ipa_version('3.0')
    assert v3.version == u'3.0'
    if hasattr(v3, '_rpmvercmp'):
        assert v3._rpmvercmp_func is None
        v3._rpmvercmp(b'1', b'2')
        assert v3._rpmvercmp_func is not None

    v4 = tasks.parse_ipa_version('4.0')
    assert v4.version == u'4.0'
    if hasattr(v4, '_rpmvercmp'):
        assert v4._rpmvercmp_func is not None

    # pylint: disable=comparison-with-itself
    assert v3 < v4
    assert v3 <= v4
    assert v3 <= v3
    assert v3 != v4
    assert v3 == v3
    assert not v3 == v4
    assert v4 > v3
    assert v4 >= v3


def test_detect_container():
    container = None
    # naive detection, may fail for OpenVZ and other container runtimes
    if os.path.isfile('/run/systemd/container'):
        with open('/run/systemd/container') as f:
            container = f.read().strip()
    elif os.geteuid() == 0:
        with open('/proc/1/environ') as f:
            environ = f.read()
        for item in environ.split('\x00'):
            if not item:
                continue
            k, v = item.split('=', 1)
            if k == 'container':
                container = v

    detected = tasks.detect_container()
    if container == 'oci':
        # systemd doesn't know about podman
        assert detected in {'container-other', container}
    else:
        assert detected == container



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from __future__ import (
    absolute_import,
    division,
)

from ipalib import (
    _,
    ngettext,
    api,
    Str,
    DNSNameParam,
    output,
    messages
)
from ipalib.errors import DependentEntry
from ipalib.plugable import Registry
from ipaserver.dns_data_management import IPASystemRecords
from ipaserver.plugins.baseldap import (
    LDAPCreate,
    LDAPSearch,
    LDAPRetrieve,
    LDAPDelete,
    LDAPObject,
    LDAPUpdate,
)
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

__doc__ = _("""
IPA locations
""") + _("""
Manipulate DNS locations
""") + _("""
EXAMPLES:
""") + _("""
  Find all locations:
    ipa location-find
""") + _("""
  Show specific location:
    ipa location-show location
""") + _("""
  Add location:
    ipa location-add location --description 'My location'
""") + _("""
  Delete location:
    ipa location-del location
""")

register = Registry()


@register()
class location(LDAPObject):
    """
    IPA locations
    """
    container_dn = api.env.container_locations
    object_name = _('location')
    object_name_plural = _('locations')
    object_class = ['top', 'ipaLocationObject']
    search_attributes = ['idnsName']
    default_attributes = [
        'idnsname', 'description'
    ]
    label = _('IPA Locations')
    label_singular = _('IPA Location')

    permission_filter_objectclasses = ['ipaLocationObject']
    managed_permissions = {
        'System: Read IPA Locations': {
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'objectclass', 'idnsname', 'description',
            },
            'default_privileges': {'DNS Administrators'},
        },
        'System: Add IPA Locations': {
            'ipapermright': {'add'},
            'default_privileges': {'DNS Administrators'},
        },
        'System: Remove IPA Locations': {
            'ipapermright': {'delete'},
            'default_privileges': {'DNS Administrators'},
        },
        'System: Modify IPA Locations': {
            'ipapermright': {'write'},
            'ipapermdefaultattr': {
                'description',
            },
            'default_privileges': {'DNS Administrators'},
        },
    }

    takes_params = (
        DNSNameParam(
            'idnsname',
            cli_name='name',
            primary_key=True,
            label=_('Location name'),
            doc=_('IPA location name'),
            # dns name must be relative, we will put it into middle of
            # location domain name for location records
            only_relative=True,
        ),
        Str(
            'description?',
            label=_('Description'),
            doc=_('IPA Location description'),
        ),
        Str(
            'servers_server*',
            label=_('Servers'),
            doc=_('Servers that belongs to the IPA location'),
            flags={'virtual_attribute', 'no_create', 'no_update', 'no_search'},
        ),
        Str(
            'dns_server*',
            label=_('Advertised by servers'),
            doc=_('List of servers which advertise the given location'),
            flags={'virtual_attribute', 'no_create', 'no_update', 'no_search'},
        ),
    )

    def get_dn(self, *keys, **options):
        loc = keys[0]
        assert isinstance(loc, DNSName)
        loc_a = loc.ToASCII()

        return super(location, self).get_dn(loc_a, **options)


@register()
class location_add(LDAPCreate):
    __doc__ = _('Add a new IPA location.')

    msg_summary = _('Added IPA location "%(value)s"')


@register()
class location_del(LDAPDelete):
    __doc__ = _('Delete an IPA location.')

    msg_summary = _('Deleted IPA location "%(value)s"')

    def pre_callback(self, ldap, dn, *keys, **options):
        assert isinstance(dn, DN)
        if not options.get('force'):
            servers = self.api.Command.server_find(
                in_location=keys[-1])['result']
            location_member = servers[0]['cn'][0] if servers else None
            if location_member:
                raise DependentEntry(
                    label=_('IPA Server'),
                    key=keys[-1],
                    dependent=location_member
                )
        system_records =IPASystemRecords(self.api)
        _success, failed = system_records.remove_location_records(keys[-1])
        if failed:
            self.add_message(messages.AutomaticDNSRecordsUpdateFailed())
        return dn


@register()
class location_mod(LDAPUpdate):
    __doc__ = _('Modify information about an IPA location.')

    msg_summary = _('Modified IPA location "%(value)s"')


@register()
class location_find(LDAPSearch):
    __doc__ = _('Search for IPA locations.')

    msg_summary = ngettext(
        '%(count)d IPA location matched',
        '%(count)d IPA locations matched', 0
    )


@register()
class location_show(LDAPRetrieve):
    __doc__ = _('Display information about an IPA location.')

    has_output = LDAPRetrieve.has_output + (
        output.Output(
            'servers',
            type=dict,
            doc=_('Servers in location'),
            flags={'no_display'},  # we use customized print to CLI
        ),
    )

    def execute(self, *keys, **options):
        result = super(location_show, self).execute(*keys, **options)

        servers_additional_info = {}
        if not options.get('raw'):
            servers_name = []
            dns_servers = []
            weight_sum = 0

            servers = self.api.Command.server_find(
                in_location=keys[0], no_members=False)['result']
            for server in servers:
                s_name = server['cn'][0]
                servers_name.append(s_name)
                weight = int(server.get('ipaserviceweight', [100])[0])
                weight_sum += weight
                servers_additional_info[s_name] = {
                    'cn': server['cn'],
                    'ipaserviceweight': server.get(
                        'ipaserviceweight', [u'100']),
                }

                s_roles = server.get('enabled_role_servrole', ())
                if s_roles:
                    servers_additional_info[s_name][
                        'enabled_role_servrole'] = s_roles
                if 'DNS server' in s_roles:
                    dns_servers.append(s_name)

            for server in servers_additional_info.values():
                server['service_relative_weight'] = [
                    u'{:.1f}%'.format(
                        int(server['ipaserviceweight'][0])*100.0/weight_sum)
                ]
            if servers_name:
                result['result']['servers_server'] = servers_name

            if dns_servers:
                result['result']['dns_server'] = dns_servers

            if not dns_servers and servers_additional_info:
                self.add_message(messages.LocationWithoutDNSServer(
                    location=keys[0]
                ))

        result['servers'] = servers_additional_info

        return result



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from ipalib import api, errors
from ipapython.dn import DN

import six

from ipatests.util import assert_deepequal, get_group_dn
from ipatests.test_xmlrpc import objectclasses
from ipatests.test_xmlrpc.xmlrpc_test import (
    fuzzy_digits, fuzzy_uuid, raises_exact)
from ipatests.test_xmlrpc.tracker.base import Tracker
from ipatests.test_xmlrpc.tracker.kerberos_aliases import KerberosAliasMixin
from ipatests.test_xmlrpc.tracker.certmapdata import CertmapdataMixin

if six.PY3:
    unicode = str


class UserTracker(CertmapdataMixin, KerberosAliasMixin, Tracker):
    """ Class for host plugin like tests """

    retrieve_keys = {
        u'dn', u'uid', u'givenname', u'sn', u'homedirectory', u'loginshell',
        u'uidnumber', u'gidnumber', u'mail', u'ou',
        u'telephonenumber', u'title', u'memberof', u'nsaccountlock',
        u'memberofindirect', u'ipauserauthtype', u'userclass',
        u'ipatokenradiusconfiglink', u'ipatokenradiususername',
        u'krbprincipalexpiration', u'usercertificate;binary',
        u'has_keytab', u'has_password', u'memberof_group', u'sshpubkeyfp',
        u'krbcanonicalname', 'krbprincipalname'
    }

    retrieve_all_keys = retrieve_keys | {
        u'usercertificate', u'street', u'postalcode',
        u'facsimiletelephonenumber', u'carlicense', u'ipasshpubkey',
        u'l', u'mobile', u'krbextradata', u'krblastpwdchange',
        u'krbpasswordexpiration', u'pager', u'st', u'manager', u'cn',
        u'ipauniqueid', u'objectclass', u'mepmanagedentry',
        u'displayname', u'gecos', u'initials', u'preserved'}

    retrieve_preserved_keys = (retrieve_keys - {u'memberof_group'}) | {
        u'preserved'}
    retrieve_preserved_all_keys = retrieve_all_keys - {u'memberof_group'}

    create_keys = retrieve_all_keys | {
        u'krbextradata', u'krbpasswordexpiration', u'krblastpwdchange',
        u'krbprincipalkey', u'userpassword', u'randompassword'}
    create_keys = create_keys - {u'nsaccountlock'}

    update_keys = retrieve_keys - {u'dn'}
    activate_keys = retrieve_keys

    find_keys = retrieve_keys - {
        u'mepmanagedentry', u'memberof_group', u'has_keytab', u'has_password',
        u'manager',
    }
    find_all_keys = retrieve_all_keys - {
        u'has_keytab', u'has_password'
    }

    primary_keys = {u'uid', u'dn'}

    def __init__(self, name=None, givenname=None, sn=None, **kwargs):
        """ Check for non-empty unicode string for the required attributes
        in the init method """

        if not (isinstance(givenname, str) and givenname):
            raise ValueError(
                "Invalid first name provided: {!r}".format(givenname)
                )
        if not (isinstance(sn, str) and sn):
            raise ValueError("Invalid second name provided: {!r}".format(sn))

        super(UserTracker, self).__init__(default_version=None)
        self.uid = unicode(name)
        self.givenname = unicode(givenname)
        self.sn = unicode(sn)
        self.dn = DN(('uid', self.uid), api.env.container_user, api.env.basedn)

        self.kwargs = kwargs

    def make_create_command(self, force=None):

        """ Make function that creates a user using user-add
            with all set of attributes and with minimal values,
            where uid is not specified """

        if self.uid is not None:
            return self.make_command(
                'user_add', self.uid,
                givenname=self.givenname,
                sn=self.sn, **self.kwargs
                )
        else:
            return self.make_command(
                'user_add', givenname=self.givenname,
                sn=self.sn, **self.kwargs
                )

    def make_delete_command(self, no_preserve=True, preserve=False):
        """ Make function that deletes a user using user-del

        Arguments 'preserve' and 'no_preserve' represent implemented
        options --preserve and --no-preserve of user-del command,
        which are mutually exclusive.
        If --preserve=True and --no-preserve=False, the user is moved
        to deleted container.
        If --preserve=True and --no-preserve=True, an error is raised.
        If --preserve=False and --no-preserver=True, user is deleted.
        """

        if preserve and not no_preserve:
            # --preserve=True and --no-preserve=False - user is moved to
            # another container, hence it is necessary to change some user
            # attributes
            self.attrs[u'dn'] = DN(
                ('uid', self.uid),
                api.env.container_deleteuser,
                api.env.basedn
                )
            self.attrs[u'objectclass'] = objectclasses.user_base

        return self.make_command(
            'user_del', self.uid,
            no_preserve=no_preserve,
            preserve=preserve
            )

    def make_retrieve_command(self, all=False, raw=False):
        """ Make function that retrieves a user using user-show """
        return self.make_command('user_show', self.uid, all=all)

    def make_find_command(self, *args, **kwargs):
        """ Make function that finds user using user-find """
        return self.make_command('user_find', *args, **kwargs)

    def make_update_command(self, updates):
        """ Make function that updates user using user-mod """
        return self.make_command('user_mod', self.uid, **updates)

    def make_undelete_command(self):
        """ Make function that activates preserved user using user-undel """
        return self.make_command('user_undel', self.uid)

    def make_enable_command(self):
        """ Make function that enables user using user-enable """
        return self.make_command('user_enable', self.uid)

    def make_disable_command(self):
        """ Make function that disables user using user-disable """
        return self.make_command('user_disable', self.uid)

    def make_stage_command(self):
        """ Make function that restores preserved user by moving it to
        staged container """
        return self.make_command('user_stage', self.uid)

    def make_group_add_member_command(self, *args, **kwargs):
        return self.make_command('group_add_member', *args, **kwargs)

    def track_create(self):
        """ Update expected state for user creation """
        self.attrs = dict(
            dn=self.dn,
            uid=[self.uid],
            givenname=[self.givenname],
            sn=[self.sn],
            homedirectory=[u'/home/%s' % self.uid],
            displayname=[u'%s %s' % (self.givenname, self.sn)],
            cn=[u'%s %s' % (self.givenname, self.sn)],
            initials=[u'%s%s' % (self.givenname[0], self.sn[0])],
            objectclass=objectclasses.user,
            description=[u'__no_upg__'],
            ipauniqueid=[fuzzy_uuid],
            uidnumber=[fuzzy_digits],
            gidnumber=[fuzzy_digits],
            krbprincipalname=[u'%s@%s' % (self.uid, self.api.env.realm)],
            krbcanonicalname=[u'%s@%s' % (self.uid, self.api.env.realm)],
            mail=[u'%s@%s' % (self.uid, self.api.env.domain)],
            gecos=[u'%s %s' % (self.givenname, self.sn)],
            loginshell=[u'/bin/sh'],
            has_keytab=False,
            has_password=False,
            mepmanagedentry=[get_group_dn(self.uid)],
            memberof_group=[u'ipausers'],
            nsaccountlock=[u'false'],
            )

        for key in self.kwargs:
            if key == u'krbprincipalname':
                try:
                    self.attrs[key] = [u'%s@%s' % (
                        (self.kwargs[key].split('@'))[0].lower(),
                        (self.kwargs[key].split('@'))[1]
                    )]
                except IndexError:
                    # we can provide just principal part
                    self.attrs[key] = [u'%s@%s' % (
                        (self.kwargs[key].lower(),
                         self.api.env.realm)
                    )]
            else:
                if type(self.kwargs[key]) is not list:
                    self.attrs[key] = [self.kwargs[key]]
                else:
                    self.attrs[key] = self.kwargs[key]

        self.exists = True

    def update(self, updates, expected_updates=None):
        """Helper function to update this user and check the result

        Overriding Tracker method for setting self.attrs correctly;
         * most attributes stores its value in list
         * the rest can be overridden by expected_updates
         * allow deleting parameters if update value is None
        """
        if expected_updates is None:
            expected_updates = {}

        self.ensure_exists()
        command = self.make_update_command(updates)
        result = command()

        for key, value in updates.items():
            if value is None or value is '' or value is u'':
                del self.attrs[key]
            elif key == 'rename':
                new_principal = u'{0}@{1}'.format(value, self.api.env.realm)
                self.attrs['uid'] = [value]
                self.attrs['krbcanonicalname'] = [new_principal]
                if new_principal not in self.attrs['krbprincipalname']:
                    self.attrs['krbprincipalname'].append(new_principal)
            else:
                if type(value) is list:
                    self.attrs[key] = value
                else:
                    self.attrs[key] = [value]
        for key, value in expected_updates.items():
            if value is None or value is '' or value is u'':
                del self.attrs[key]
            else:
                self.attrs[key] = value

        self.check_update(
            result,
            extra_keys=set(updates.keys()) | set(expected_updates.keys())
        )

        if 'rename' in updates:
            self.uid = self.attrs['uid'][0]

    def check_create(self, result, extra_keys=()):
        """ Check 'user-add' command result """
        expected = self.filter_attrs(self.create_keys | set(extra_keys))
        assert_deepequal(dict(
            value=self.uid,
            summary=u'Added user "%s"' % self.uid,
            result=self.filter_attrs(expected),
            ), result)

    def check_delete(self, result):
        """ Check 'user-del' command result """
        assert_deepequal(dict(
            value=[self.uid],
            summary=u'Deleted user "%s"' % self.uid,
            result=dict(failed=[]),
            ), result)

    def check_retrieve(self, result, all=False, raw=False):
        """ Check 'user-show' command result """
        if u'preserved' in self.attrs and self.attrs[u'preserved']:
            self.retrieve_all_keys = self.retrieve_preserved_all_keys
            self.retrieve_keys = self.retrieve_preserved_keys
        elif u'preserved' not in self.attrs and all:
            self.attrs[u'preserved'] = False

        if all:
            expected = self.filter_attrs(self.retrieve_all_keys)
        else:
            expected = self.filter_attrs(self.retrieve_keys)

        # small override because stageuser-find returns different type
        # of nsaccountlock value than DS, but overall the value fits
        # expected result
        if u'nsaccountlock' in expected:
            if expected[u'nsaccountlock'] == [u'true']:
                expected[u'nsaccountlock'] = True
            elif expected[u'nsaccountlock'] == [u'false']:
                expected[u'nsaccountlock'] = False

        assert_deepequal(dict(
            value=self.uid,
            summary=None,
            result=expected,
        ), result)

    def check_find(self, result, all=False, pkey_only=False, raw=False,
                   expected_override=None):
        """ Check 'user-find' command result """
        if all:
            if u'preserved' not in self.attrs:
                self.attrs.update(preserved=False)
            expected = self.filter_attrs(self.find_all_keys)
        elif pkey_only:
            expected = self.filter_attrs(self.primary_keys)
        else:
            expected = self.filter_attrs(self.find_keys)

        if all and self.attrs[u'preserved']:
            del expected[u'mepmanagedentry']

        if u'nsaccountlock' in expected:
            if expected[u'nsaccountlock'] == [u'true']:
                expected[u'nsaccountlock'] = True
            elif expected[u'nsaccountlock'] == [u'false']:
                expected[u'nsaccountlock'] = False

        if expected_override:
            assert isinstance(expected_override, dict)
            expected.update(expected_override)

        assert_deepequal(dict(
            count=1,
            truncated=False,
            summary=u'1 user matched',
            result=[expected],
        ), result)

    def check_find_nomatch(self, result):
        """ Check 'user-find' command result when no user should be found """
        assert_deepequal(dict(
            count=0,
            truncated=False,
            summary=u'0 users matched',
            result=[],
        ), result)

    def check_update(self, result, extra_keys=()):
        """ Check 'user-mod' command result """
        expected = self.filter_attrs(self.update_keys | set(extra_keys))
        if expected[u'nsaccountlock'] == [u'true']:
            expected[u'nsaccountlock'] = True
        elif expected[u'nsaccountlock'] == [u'false']:
            expected[u'nsaccountlock'] = False

        assert_deepequal(dict(
            value=self.uid,
            summary=u'Modified user "%s"' % self.uid,
            result=expected
        ), result)

    def check_enable(self, result):
        """ Check result of enable user operation """
        assert_deepequal(dict(
            value=self.name,
            summary=u'Enabled user account "%s"' % self.name,
            result=True
        ), result)

    def check_disable(self, result):
        """ Check result of disable user operation """
        assert_deepequal(dict(
            value=self.name,
            summary=u'Disabled user account "%s"' % self.name,
            result=True
        ), result)

    def create_from_staged(self, stageduser):
        """ Copies attributes from staged user - helper function for
        activation tests """
        self.attrs = stageduser.attrs
        self.uid = stageduser.uid
        self.givenname = stageduser.givenname
        self.sn = stageduser.sn

        self.attrs[u'mepmanagedentry'] = None
        self.attrs[u'dn'] = self.dn
        self.attrs[u'ipauniqueid'] = [fuzzy_uuid]
        self.attrs[u'memberof_group'] = [u'ipausers']
        self.attrs[u'mepmanagedentry'] = [u'cn=%s,%s,%s' % (
            self.uid, api.env.container_group, api.env.basedn
            )]
        self.attrs[u'objectclass'] = objectclasses.user
        if self.attrs[u'gidnumber'] == [u'-1']:
            self.attrs[u'gidnumber'] = [fuzzy_digits]
        if self.attrs[u'uidnumber'] == [u'-1']:
            self.attrs[u'uidnumber'] = [fuzzy_digits]

        if u'ipasshpubkey' in self.kwargs:
                self.attrs[u'ipasshpubkey'] = [str(
                    self.kwargs[u'ipasshpubkey']
                    )]
        self.attrs[u'nsaccountlock'] = [u'false']

    def check_activate(self, result):
        """ Check 'stageuser-activate' command result """
        expected = dict(
            value=self.uid,
            summary=u'Stage user %s activated' % self.uid,
            result=self.filter_attrs(self.activate_keys))

        # small override because stageuser-find returns different
        # type of nsaccountlock value than DS, but overall the value
        # fits expected result
        if expected['result'][u'nsaccountlock'] == [u'true']:
            expected['result'][u'nsaccountlock'] = True
        elif expected['result'][u'nsaccountlock'] == [u'false']:
            expected['result'][u'nsaccountlock'] = False

        assert_deepequal(expected, result)

        self.exists = True

    def check_undel(self, result):
        """ Check 'user-undel' command result """
        assert_deepequal(dict(
            value=self.uid,
            summary=u'Undeleted user account "%s"' % self.uid,
            result=True
            ), result)

    def enable(self):
        """ Enable user account if it was disabled """
        if (self.attrs['nsaccountlock'] is True or
                self.attrs['nsaccountlock'] == [u'true']):
            self.attrs.update(nsaccountlock=False)
            result = self.make_enable_command()()
            self.check_enable(result)

    def disable(self):
        """ Disable user account if it was enabled """
        if (self.attrs['nsaccountlock'] is False or
                self.attrs['nsaccountlock'] == [u'false']):
            self.attrs.update(nsaccountlock=True)
            result = self.make_disable_command()()
            self.check_disable(result)

    def track_delete(self, preserve=False):
        """Update expected state for host deletion"""
        if preserve:
            self.exists = True
            if u'memberof_group' in self.attrs:
                del self.attrs[u'memberof_group']
            self.attrs[u'nsaccountlock'] = True
            self.attrs[u'preserved'] = True
        else:
            self.exists = False
            self.attrs = {}

    def make_preserved_user(self):
        """ 'Creates' a preserved user necessary for some tests """
        self.ensure_exists()
        self.track_delete(preserve=True)
        command = self.make_delete_command(no_preserve=False, preserve=True)
        result = command()
        self.check_delete(result)

    def check_attr_preservation(self, expected):
        """ Verifies that ipaUniqueID, uidNumber and gidNumber are
        preserved upon reactivation. Also verifies that resulting
        active user is a member of ipausers group only."""
        command = self.make_retrieve_command(all=True)
        result = command()

        assert_deepequal(dict(
            ipauniqueid=result[u'result'][u'ipauniqueid'],
            uidnumber=result[u'result'][u'uidnumber'],
            gidnumber=result[u'result'][u'gidnumber']
            ), expected)

        if (u'memberof_group' not in result[u'result'] or
                result[u'result'][u'memberof_group'] != (u'ipausers',)):
            assert False

    def make_fixture_restore(self, request):
        """Make a pytest fixture for a preserved user that is to be moved to
        staged area.

        The fixture ensures the plugin entry does not exist before
        and after the tests that use it. It takes into account
        that the preserved user no longer exists after restoring it,
        therefore the fixture verifies after the tests
        that the preserved user doesn't exist instead of deleting it.
        """
        del_command = self.make_delete_command()
        try:
            del_command()
        except errors.NotFound:
            pass

        def finish():
            with raises_exact(errors.NotFound(
                    reason=u'%s: user not found' % self.uid)):
                del_command()

        request.addfinalizer(finish)

        return self

    def make_admin(self, admin_group=u'admins'):
        """ Add user to the administrator's group """
        result = self.run_command('group_show', admin_group)
        admin_group_content = result[u'result'][u'member_user']
        admin_group_expected = list(admin_group_content) + [self.name]

        command = self.make_group_add_member_command(
            admin_group, **dict(user=self.name)
        )
        result = command()
        assert_deepequal(dict(
            completed=1,
            failed=dict(
                member=dict(group=tuple(), user=tuple(), service=tuple())
            ),
            result={
                'dn': get_group_dn(admin_group),
                'member_user': admin_group_expected,
                'gidnumber': [fuzzy_digits],
                'cn': [admin_group],
                'description': [u'Account administrators group'],
            },
        ), result)

    #  Kerberos aliases methods
    def _make_add_alias_cmd(self):
        return self.make_command('user_add_principal', self.name)

    def _make_remove_alias_cmd(self):
        return self.make_command('user_remove_principal', self.name)

    # Certificate identity mapping methods
    def _make_add_certmap(self):
        return self.make_command('user_add_certmapdata', self.name)

    def _make_remove_certmap(self):
        return self.make_command('user_remove_certmapdata', self.name)
